import os, pkg_resources, shutil


def create_tutorials(destination, silent = False):
    """
    Copies mercury.graph tutorial notebooks to `destination`. A folder will be created inside
    output_dir, named 'graph_tutorials'. The folder `destination` must exist.

    Args:
        destination (str): The destination directory
        silent (bool): If True, suppresses output on success.

    Raises:
        ValueError: If `output_dir` is equal to source path.

    Examples:
        >>> # copy tutorials to /tmp/graph_tutorials
        >>> from mercury.graph import create_tutorials
        >>> create_tutorials('/tmp')

    """
    src = pkg_resources.resource_filename(__package__, 'tutorials')
    dst = os.path.abspath(destination)

    assert src != dst, 'Destination (%s) cannot be the same as source.' % src

    assert os.path.isdir(dst), 'Destination (%s) must be a directory.' % dst

    dst = os.path.join(dst, 'graph_tutorials')

    assert not os.path.exists(dst), 'Destination (%s) already exists' % dst

    shutil.copytree(src, dst)

    if not silent:
        print('Tutorials copied to: %s' % dst)



import numpy as np

from scipy.spatial.distance import cdist
from mercury.graph.core._njit import njit

from mercury.graph.core.base import BaseClass


@njit
def _elliptic_rotate(self_em, iu, iv, cos_w, sin_w):
    for i in range(len(iu)):
        j = iu[i]

        u = self_em[j]
        v = self_em[iv[i]]

        sc = np.dot(u, v) / np.dot(u, u)
        pv = sc * u
        tv = v - pv

        self_em[j] = cos_w * pv / sc + sin_w * tv

    return self_em


class Embeddings(BaseClass):
    """
    This class holds a matrix object that is interpreted as the embeddings for any list of objects, not only the nodes of a graph. You
    can see this class as the internal object holding the embedding for other classes such as class GraphEmbedding.

    Args:
        dimension (int): The number of columns in the embedding. See note below.
        num_elements (int): The number of rows in the embedding. You can leave this empty on creation and then use initialize_as() to
            automatically match the nodes in a graph.
        mean (float): The (expected) mean of the initial values.
        sd (float): The (expected) standard deviation of the initial values.
        learn_step (float): The size of the learning step elements get approached or moved away. Units are hexadecimal degrees in along
            an ellipse.
        bidirectional (bool): Should the changes apply only to the elements of first column (False) or to both.

    Note:
        **On dimension:** Embeddings cannot be zero (that is against the whole concept). Smaller dimension embeddings can only hold
        few elements without introducing spurious correlations by some form of 'birthday attack' phenomenon as elements increase. Later
        it is very hard to get rid of that spurious 'knowledge'. 

        **Solution**: With may elements, you have to go to high enough dimension even if the structure is simple. 
        Pretending to fit many embeddings in low dimension without them being correlated is like pretending to plot a trillion random 
        points in a square centimeter while keeping them 1 mm apart from each other: It's simply impossible!
    """

    def __init__(
        self, dimension, num_elements=0, mean=0, sd=1, learn_step=3, bidirectional=False
    ):
        self.dimension = dimension
        self.num_elements = num_elements
        self.mean = mean
        self.sd = sd
        self.learn_step = learn_step
        self.bidirectional = bidirectional

        if self.num_elements > 0:
            self.embeddings_matrix_ = np.random.normal(
                self.mean, self.sd, (self.num_elements, self.dimension)
            )

    def fit(self, converge=None, diverge=None):
        """
        Apply a learning step to the embedding.

        Args:
            converge (numpy matrix of two columns): A matrix of indices to elements meaning (first column) should be approached to
                (second column).
            diverge (numpy matrix of two columns): A matrix of indices to elements meaning (first column) should be moved away from
                (second column).

        Returns:
            (self): Fitted self (or raises an error)

        Note:
            Embeddings start being randomly distributed and hold no structure other than spurious correlations. Each time you apply a
            learning step by calling this method, you are tweaking the embedding to approach some rows and/or move others away. You can use
            both converge and diverge or just one of them and call this as many times you want with varying learning step. A proxy of how
            much an embedding can learn can be estimated by measuring how row correlations are converging towards some asymptotic values.
        """

        w = self.learn_step * np.pi / 180

        cos_w = np.cos(w)
        sin_w = np.sin(w)

        if converge is not None:
            self.embeddings_matrix_ = _elliptic_rotate(
                self.embeddings_matrix_, converge[:, 0], converge[:, 1], cos_w, sin_w
            )

            if self.bidirectional:
                self.embeddings_matrix_ = _elliptic_rotate(
                    self.embeddings_matrix_,
                    converge[:, 1],
                    converge[:, 0],
                    cos_w,
                    sin_w,
                )

        if diverge is not None:
            self.embeddings_matrix_ = _elliptic_rotate(
                self.embeddings_matrix_, diverge[:, 0], diverge[:, 1], cos_w, -sin_w
            )

            if self.bidirectional:
                self.embeddings_matrix_ = _elliptic_rotate(
                    self.embeddings_matrix_, diverge[:, 1], diverge[:, 0], cos_w, -sin_w
                )

        return self

    def as_numpy(self):
        """
        Return the embedding as a numpy matrix where each row is an embedding.
        """
        if not hasattr(self, "embeddings_matrix_"):
            return

        return self.embeddings_matrix_

    def get_most_similar_embeddings(self, index, k=5, metric="cosine"):
        """
        Given an index of a vector in the embedding matrix, returns the k most similar embeddings in the matrix

        Args:
            index (int): index of the vector in the matrix that we want to compute the similar embeddings
            k (int): Number of most similar embeddings to return
            metric (str): metric to use as a similarity.

        Returns:
            (list): list of k most similar nodes as indices and list of similarities of the most similar nodes
        """
        if metric == "cosine":
            similarities = (
                1
                - cdist(
                    np.expand_dims(self.as_numpy()[index], axis=0),
                    self.as_numpy(),
                    "cosine",
                )[0]
            )

        elif metric == "euclidean":
            similarities = 1 / (
                1
                + cdist(
                    np.expand_dims(self.as_numpy()[index], axis=0),
                    self.as_numpy(),
                    "euclidean",
                )[0]
            )

        else:
            raise ValueError("Unknown Distance Metric: %s" % metric)

        ordered_indices = np.argsort(similarities)[::-1][1 : (k + 1)]
        ordered_similarities = similarities[ordered_indices]

        return ordered_indices, ordered_similarities



import warnings

import pandas as pd
import networkx as nx

from numpy.linalg import matrix_power

from mercury.graph.core import Graph
from mercury.graph.core.base import BaseClass


class Transition(BaseClass):
    """
    Create an interface class to manage the adjacency matrix of a directed graph as a transition matrix.
    This enables computing distributions of probabilities over the nodes after a given number of iterations.
    """

    def __init__(self):
        self.fitted_graph_ = None

    def fit(self, G: Graph):
        """
        Converts the adjacency matrix into a transition matrix. Transition matrices are used to compute the distribution of probability
        of being in each of the nodes (or states) of a directed graph (or Markov process). The distribution for state s is:

        * $s_t = T*s_{t-1}$

        Where:

        T is the transition matrix. After calling.fit(), the adjacency matrix is the transition matrix. You can use .to_pandas() to see it.
        $s_{t-1}$ is the previous state.

        What .fit() does is scaling the non-zero rows to make them sum 1 as they are probability distributions and make the zero rows
        recurrent states. A recurrent state is a final state, a state whose next state is itself.

        Args:
            G (Graph): A `mercury.graph` Graph.

        Returns:
            (self): Fitted self (or raises an error).

        Note:
            If created using NetworkX directly, the name of the weight must be 'weight' and must be positive. The recommended way
            to create the graph is using .set_row() which will always name the weight as 'weight' but does not check the value.

        """
        names = list(G.networkx.nodes)
        adj_m = nx.adjacency_matrix(G.networkx, weight="weight", dtype=float)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")

            for i in range(adj_m.shape[0]):
                row = adj_m[[i], :]
                tot = row.sum()

                if tot == 0:
                    row[0, i] = 1
                else:
                    row = row / tot

                adj_m[[i], :] = row

        df = pd.DataFrame(adj_m.todense(), index=names, columns=names)
        self.fitted_graph_ = Graph(nx.from_pandas_adjacency(df, create_using=nx.DiGraph))

        return self

    def to_pandas(self, num_iterations=1):
        """
        Returns the adjacency (which is the transition matrix after `fit()` was called) for a given number of iterations as a pandas
        dataframe with labeled rows and columns.

        Args:
            num_iterations (int): If you want to compute the matrix for a different number of iterations, k, you can use this argument to
                raise the matrix to any non negative integer, since $s_{t+k} = T^k*s_t$

        Returns:
            (pd.DataFrame): The transition matrix for num_iterations.

        Note:
            This method does not automatically call `fit()`. This allows inspecting the adjacency matrix as a pandas dataframe.
            The result of computing num_iterations will not make sense if `fit()` has not been called before `to_pandas()`.

        """
        if self.fitted_graph_ is None:
            raise ValueError("Error: fit() must be called first.")

        names = list(self.fitted_graph_.networkx.nodes)
        adj_m = nx.adjacency_matrix(self.fitted_graph_.networkx, weight="weight").todense()

        if num_iterations != 1:
            adj_m = matrix_power(adj_m, num_iterations)

        return pd.DataFrame(adj_m, index=names, columns=names)



# Imports
import warnings
from itertools import product
from typing import Union, List
from mercury.graph.core.base import BaseClass
from mercury.graph.core import Graph
from mercury.graph.core.spark_interface import pyspark_installed

# Pyspark imports
if pyspark_installed:
    from pyspark.sql import (
        functions as F,
        Window as W,
        DataFrame,
        SparkSession
    )
    from pyspark.sql.types import (
        ByteType, ShortType, IntegerType, LongType,
        FloatType, DoubleType, DecimalType
    )


# Main class to be fitted
class GraphFeatures(BaseClass):
    """Generate features from neighboring nodes.

        Args:
            attributes (str or List[str], optional):
                The node attributes used to generate features. These strings
                must be valid column names in the graph's vertices'. If None,
                all columns except for 'id' are used. Default is None.
            agg_funcs (str or List[str], optional):
                The aggregation function(s) to apply. Supported values are:
                - "sum"
                - "min"
                - "max"
                - "avg"
                - "wavg"
                Default is "avg".
            order (int, optional):
                The order of neighbors to compute (e.g., 1 for immediate
                neighbors, 2 for neighbors of neighbors, etc.). It must be a
                positive integer. Default is 1.
            verify (bool, optional):
                Whether to validate the provided parameters before executing
                the algorithm. Default is False.
            checkpoint (bool, optional):
                Whether to use Spark checkpointing to persist intermediate
                results. Default is False.
            checkpoint_dir (str, optional):
                The directory to store checkpoint files if checkpointing is
                enabled. Required if `checkpoint=True`. Default is None.
            spark (SparkSession, optional):
                The active Spark session. Required if `checkpoint=True`.
                Default is None.

        Notes:
            - The `order` parameter specifies the depth of the neighborhood
            considered  during aggregation.
            - If `wavg` is specified in `agg_funcs`, the `edges` DataFrame must
            include a 'weight' column. Weighted averages are computed as the
            product of attribute values and edge weights, normalized by the
            total weight.
            - Loops back to oneself are excluded (e.g., 0 -> 1 -> 0).

        Examples:
            >>> g = Graph(data=edges, nodes=vertices, keys={"directed": False})
            >>> gf = GraphFeatures(
            ...     attributes=["feature1", "feature2"],  # Valid names in `vertices`
            ...     agg_funcs=["min", "max", "avg"]
            ... )
            >>> gf.fit(g)  # Adds `node_features_` attribute to `gf`
            >>> gf.node_features_.show(5)  # Display new graph features
        """

    def __init__(
            self,
            attributes: Union[str, List[str]] = None,
            agg_funcs: Union[str, List[str]] = 'avg',
            order: int = 1,
            verify: bool = False,
            checkpoint: bool = False,
            checkpoint_dir: str = None,
            spark: SparkSession = None,
    ):
        self.attributes = attributes
        self.agg_funcs = agg_funcs
        self.order = order
        self.verify = verify
        self.checkpoint = checkpoint
        self.checkpoint_dir = checkpoint_dir
        self.spark = spark

    def __str__(self):
        pass

    # Aggregate messages from neighbors
    def fit(self, g: Graph):
        """Fit the GraphFeatures object to generate graph-based features.

        Args:
            g (Graph): A mercury graph object (mercury.graph.core.graph.Graph).

        Returns:
            (self): Fitted self with 'node_features_' attribute (or raises an
                error).
        """

        # Get parameters
        edges = g.graphframe.edges
        vertices = g.graphframe.vertices
        attributes = self.attributes
        agg_funcs = self.agg_funcs
        order = self.order
        verify = self.verify
        checkpoint = self.checkpoint
        checkpoint_dir = self.checkpoint_dir
        spark = self.spark

        # Dictionary that maps agg_funcs to sql functions
        _agg_dict = {
            'sum': F.sum,
            'min': F.min,
            'max': F.max,
            'avg': F.avg,
            'wavg': F.sum
        }

        # Verify attributes
        if attributes is None:
            attributes = [col for col in vertices.columns if col != 'id']
        elif isinstance(attributes, str):
            attributes = [attributes]
        else:
            msg = 'attributes must be a list of strings.'
            assert isinstance(attributes, list), msg
            msg = 'All items in attributes must be strings.'
            assert all(isinstance(item, str) for item in attributes), msg
            msg = 'All items in attributes must exist in vertices.'
            assert all(item in vertices.columns for item in attributes), msg
            msg = 'attributes must only contain attributes ("id" was detected).'
            assert 'id' not in attributes, msg

        # Verify aggregation functions
        agg_funcs = [agg_funcs] if isinstance(agg_funcs, str) else agg_funcs
        msg = (
            'Invalid aggregation function. Please provide one or more of the following '
            'valid options: "sum", "min", "max", "avg" or "wavg".'
        )
        assert all([func in _agg_dict.keys() for func in agg_funcs]), msg

        # Verify vertices
        verify and self._verify_vertices(vertices)

        # Verify edges
        verify and self._verify_edges(edges)

        # Verify checkpoint configuration and Spark session integrity
        if checkpoint:
            assert checkpoint_dir is not None, (
                'checkpoint_dir must be provided when checkpoint is True.'
            )
            assert spark is not None, (
                'spark must be provided when checkpoint is True.'
            )
            assert isinstance(checkpoint_dir, str), (
                'Invalid type for checkpoint_dir: expected str (a valid path), but got'
                f' {type(checkpoint_dir).__name__} instead.'
            )
            assert isinstance(spark, SparkSession), (
                'Invalid type for spark: expected SparkSession instance, but got '
                f'{type(spark).__name__} instead.'
            )

        # Fetch neighbors
        neighs = self._get_neighbors(
            edges,
            order,
            checkpoint,
            checkpoint_dir,
            spark
        )

        # Fetch neighbors' attributes
        ret = (
            neighs
            .join(
                other=(
                    vertices
                    .select(
                        F.col('id').alias('neigh'),
                        *attributes
                    )
                ),
                on='neigh',
                how='left'
            )
        )

        # Temporarily assign columns if user is calculating weighted averages
        if 'wavg' in agg_funcs:
            ret = ret.select(
                F.expr('*'),
                *[(F.col(x) * F.col('weight')).alias('w_' + x)
                  for x in attributes]
            )

        # Combine functions and attributes
        _agg = list(product(agg_funcs, attributes))

        # Apply all functions to all attributes
        ret = (
            ret
            .groupBy('id')
            .agg(
                *[
                    _agg_dict[k](x).alias('_'.join([x, k]))
                    for (k, x) in _agg if k != 'wavg'
                ],
                *[
                    _agg_dict[k]('w_' + x).alias('_'.join([x, 'wavg']))
                    for (k, x) in _agg if k == 'wavg'
                ]
            )
        )

        # Return result
        self.node_features_ = ret
        return self

    # Check vertices integrity
    def _verify_vertices(self, vertices: DataFrame):
        """Validates the integrity of the 'vertices' DataFrame for graph
           processing. This function verifies the presence of an 'id' column,
           ensures that the 'id' column contains unique values, checks that
           all other columns contain numeric data types only, and identifies
           any null values in non-'id' columns.

        Args:
            vertices : DataFrame
                A PySpark DataFrame representing vertices in a graph.

        Returns:
            None
        """

        # Check type
        msg = 'vertices must be a PySpark DataFrame.'
        assert isinstance(vertices, DataFrame), msg

        # Check if id column exists
        msg = "Column 'id' not listed in vertices."
        assert 'id' in vertices.columns, msg

        # Assert unique IDs
        ids_duplicated = (
            vertices
            .select(
                F.count('id').over(
                    W.partitionBy('id')
                ).alias('count')
            )
            .where(F.col('count') > 1)
            .limit(1)
            .count()
        )
        msg = 'vertices contains duplicated IDs.'
        assert ids_duplicated == 0, msg

        # Numeric Types
        numeric_types = (
            ByteType, ShortType,
            IntegerType, LongType,
            FloatType, DoubleType, DecimalType
        )

        # Check for nulls (dict where key=col_name and value=null_count)
        nulls_dict = (
            vertices
            .select(
                *[
                    F.count(F.when(F.col(c).isNull(), c)).alias(c)
                    for c in vertices.columns if c != 'id'
                ]
            )
            .collect()[0]
            .asDict()
        )

        # Check for nulls and data types every column in the DataFrame
        for field in vertices.schema.fields:
            col_name = field.name

            # Skip column with IDs
            if col_name == 'id':
                continue
            col_type = type(field.dataType)
            col_nulls = nulls_dict[col_name]

            # Data type check
            assert (
                col_type in numeric_types
            ), f"Non-numeric values found in column '{col_name}'."

            # Nulls check
            if col_nulls > 0:
                warnings.warn(
                    f"Column '{col_name}' contains {col_nulls} null values"
                )

    # Check edges' integrity
    def _verify_edges(self, edges: DataFrame):
        """Validates the integrity of the 'edges' DataFrame for graph
           processing. This function verifies the presence of 'src' and 'dst'
           columns, ensures that each pair (src, dst) is unique, confirms the
           graph is undirected, and if a 'weight' column exists, verifies it
           has a numeric data type.

        Args:
            edges : DataFrame
                A PySpark DataFrame representing edges in a graph.

        Returns:
            None
        """

        # Check type
        msg = 'edges must be a PySpark DataFrame.'
        assert isinstance(edges, DataFrame), msg

        # Assert edges
        msg = 'Expected column "{}" not in edges'
        assert 'src' in edges.columns, msg.format('src')
        assert 'dst' in edges.columns, msg.format('dst')

        # Assert unique pairs
        pairs_duplicated = (
            edges
            .groupBy('src', 'dst')
            .count()
            .where(F.col('count') > 1)
            .count()
        )
        msg = 'edges contains {} duplicated src-dst pairs.'
        assert pairs_duplicated == 0, msg.format(pairs_duplicated)

        # Assert undirected
        pairs_mirrored = (
            edges
            .select('src', 'dst')
            .unionByName(
                edges
                .where(F.col('src') != F.col('dst'))  # Drop self-loops
                .select(
                    F.col('dst').alias('src'),
                    F.col('src').alias('dst')
                )
            )
            .groupBy('src', 'dst')
            .count()
            .where(F.col('count') > 1)
            .count()
        )
        msg = 'edges has mirrored edges. Directed graphs are not yet supported!'
        assert pairs_mirrored == 0, msg

        # Assert weight dtype
        if 'weight' in edges.columns:
            msg = 'Column "weight" must be a float or an int. Received {} instead.'
            dtype = dict(edges.dtypes)['weight']
            assert dtype in (
                'float', 'double', 'tinyint', 'smallint', 'int', 'bigint'
            ), msg.format(dtype)

    # Fetches all the neighbors of each node in edges
    def _get_neighbors(
        self,
        edges: DataFrame,
        order: int = 1,
        checkpoint: bool = False,
        checkpoint_dir: str = None,
        spark: SparkSession = None,
    ):
        """Fetch all the neighbors of each node in a graph up to specific
           order.

        Args:
            edges : DataFrame
                A Spark DataFrame containing the edges of the graph. It must
                contain the columns 'src' and 'dst' corresponding to source
                and destination nodes, respectively. An optional 'weight'
                column can be included; if not present, it is assumed to be 1.
            order : int, optional
                The order of neighbors to compute (e.g., 1 for immediate
                neighbors, 2 for neighbors of neighbors, etc.). It must be a
                positive integer. Default is 1.
            checkpoint : bool, optional
                Whether to use Spark checkpointing to persist intermediate
                results. Default is False.
            checkpoint_dir : str, optional,
                The directory to store checkpoint files if checkpointing is
                enabled. Required if `checkpoint=True`. Default is None.
            spark : SparkSession, optional
                The active Spark session. Required if `checkpoint=True`.
                Default is None.

        Returns:
            DataFrame
                A Spark DataFrame with the columns:
                - 'id': The node ID.
                - 'neigh': The ID of a neighboring node.
                - 'weight': The computed weight of the edge connecting 'id' to
                'neigh', normalized by the total weight of all edges
                originating from 'id'.

        Notes:
            - Self-loops are removed when calculating weighted aggregations.
        """
        # Add weight column if necessary
        if 'weight' not in edges.columns:
            edges = edges.withColumn('weight', F.lit(1))

        # Assert order
        assert isinstance(order, int), 'order must be an integer.'
        assert order > 0, 'order must be an integer greater than 0.'
        if order > 2:
            warnings.warn(
                f"order={order} may cause the process to be slow."
            )

        # Fetch 1st-degree neighbors (preserve self-loops for weighted degree)
        ret = (
            edges
            .unionByName(
                edges
                .select(
                    F.col('dst').alias('src'),
                    F.col('src').alias('dst'),
                    'weight'
                )
            )
            # Get weights (accounting for self-loops)
            .select(
                F.col('src').alias('id'),
                F.col('dst').alias('neigh'),
                (
                    F.col('weight')
                    / F.sum('weight').over(
                        W.partitionBy(F.col('src'))
                    )
                ).alias('weight')
            )
            # Drop self-loops AFTER calculating weights
            .where(F.col('id') != F.col('neigh'))
        )

        # checkpoint result
        if checkpoint:
            spark.sparkContext.setCheckpointDir(checkpoint_dir)
            ret = ret.checkpoint()

        # Fetch n-degree neighbors
        for _ in range(1, order):
            ret_tmp = (
                ret.select(
                    F.col('id'),
                    F.col('neigh').alias('o1_neigh'),
                    F.col('weight').alias('o1_weight')
                )
                .join(
                    other=ret.select(
                        F.col('id').alias('o1_neigh'),
                        F.col('neigh').alias('o2_neigh'),
                        F.col('weight').alias('o2_weight')
                    ),
                    on='o1_neigh',
                    how='left'
                )
                # Drop paths back to self
                .where(F.col('id') != F.col('o2_neigh'))
                .select(
                    F.col('id'),
                    F.col('o2_neigh').alias('neigh'),
                    (
                        F.col('o1_weight') * F.col('o2_weight')
                    ).alias('weight')
                )
            )
            # checkpoint result
            if checkpoint:
                ret = ret_tmp.checkpoint()
            else:
                ret = ret_tmp
        # Return
        return ret



import importlib.util, json, os

import anywidget
import pathlib
import traitlets

import pandas as pd

from mercury.graph.core.spark_interface import SparkInterface

display, HTML = None, None
if importlib.util.find_spec('IPython.display') is not None:
    from IPython.display import display, HTML


class MoebiusAnywidget(anywidget.AnyWidget):
    """
    MoebiusAnywidget class for visualizing graphs creating an anywidget that uses JavaScript and HTML.
    Note: It is recommended to use wrapper class Moebius.

    **Important**: This class requires: anywidget, traitlets and IPython.display!! These packages are not mandatory
    for the rest of the library, since you may not be intereseted in visualizing graphs. If you want to use this
    class, you need to install these packages.

    Args:
        G (Graph):          The graph to be visualized.
        initial_id (str): The id of the node to start the visualization.
        initial_depth (int): The initial depth of the graph (starting with `initial_id` as 0) to be shown.
        node_config (dict): A node configuration dictionary created by `node_config()`.
        edge_config (dict): An edge configuration dictionary created by `edge_config()`.

    Attributes:
        use_spark (bool):   Flag indicating if Spark is used.
        front_pat (str):    Path to the frontend resources.
        _int_id_map (dict): A dictionary mapping node IDs to integer IDs.
    """

    def __init__(self, G, initial_id = None, initial_depth = 1, node_config = None, edge_config = None):
        if display is None or HTML is None:
            raise ImportError('IPython is not installed')

        self.G = G
        self.use_spark = self.G._as_networkx is None
        self.front_pat = os.path.dirname(os.path.dirname(os.path.realpath(__file__))) + '/frontend'
        self._int_id_map = {node['id'] : i for i, node in enumerate(self.G.nodes)}

        if initial_id is None:
            initial_id = next(iter(self._int_id_map))

        initial_json = self._get_adjacent_nodes_moebius(initial_id, depth = initial_depth)

        if node_config is None:
            node_config = self.node_or_edge_config()

        if edge_config is None:
            edge_config = self.node_or_edge_config()

        logo_svg = "`" + pathlib.Path(self.front_pat + '/moebius.svg.html').read_text() + "`"

        js_moebius_call = f"""
            (function(el) {{
                moebius({initial_json}, {node_config}, {edge_config}, {str(self.G.is_directed).lower()}, {logo_svg});
            }})(el);
        """

        self._esm = f"""
            import {{ min, max }} from 'https://esm.sh/d3-array@3';
            import {{ drag }} from 'https://esm.sh/d3-drag@3';
            import {{ forceX, forceY, forceCollide, forceSimulation, forceLink, forceManyBody, forceCenter }} from 'https://esm.sh/d3-force@3';
            import {{ scaleOrdinal, scaleLinear, scalePow, scaleSqrt, scaleLog }} from 'https://esm.sh/d3-scale@4';
            import {{ schemeCategory10 }} from 'https://esm.sh/d3-scale-chromatic@3';
            import {{ select, selectAll }} from 'https://esm.sh/d3-selection@3';
            import {{ transition }} from 'https://esm.sh/d3-transition@3';
            import {{ zoom }} from 'https://esm.sh/d3-zoom@3';
            const d3 = {{ min, max, select, selectAll, forceSimulation, forceLink, forceManyBody, forceCenter, drag, scaleOrdinal, schemeCategory10, scaleOrdinal, scaleLinear, scalePow, scaleSqrt, scaleLog, zoom, forceX, forceY, forceCollide, transition }};

            export default {{
                initialize({{ model }}) {{
                    // Set up shared state or event handlers.
                    return () => {{
                    // Optional: Called when the widget is destroyed.
                    }}
                }},
                render({{ model, el }}) {{
                    // Render the widget's view into the el HTMLElement.
                    {pathlib.Path(self.front_pat + '/moebius.js').read_text()}
                    {js_moebius_call}
                    return () => {{
                    // Optional: Called when the view is destroyed.
                    }}
                }}
            }}
            """

        self._css = pathlib.Path(self.front_pat + '/moebius.css').read_text()

        super().__init__()


    expandNode_params = traitlets.Dict({}).tag(sync=True)
    searchNewNode_params = traitlets.Dict({}).tag(sync=True)

    # Observers that trigger Python code execution when frontend registers calls to expandNode and searchNewNode
    @traitlets.observe("expandNode_params")
    def _on_change_expandNode_params(self, change):
        try:
            my_value = self._get_adjacent_nodes_moebius(change["new"]["nodeID"], change["new"]["nodeLimit"], change["new"]["depth"])
            self.send({"result": my_value})
        except:
            self.send({"result": ""})

    @traitlets.observe("searchNewNode_params")
    def _on_change_searchNewNode_params(self, change):
        try:
            my_value = self._get_adjacent_nodes_moebius(change["new"]["nodeID"], change["new"]["nodeLimit"], change["new"]["depth"])
            self.send({"result": my_value, "searched_nodeID": change["new"]["nodeID"]})
        except:
            self.send({"result": ""})


    def __str__(self):
        """
        Convert the object via str()
        """

        return 'Moebius(%s)' % str(self.G)


    def __getitem__(self, item):
        """
        Add support for the [] operator.
        """

        return self._get_adjacent_nodes_moebius(item)


    def generate_color_palette(self, cats, hue = 0, sat = 0.7, light = 0.5):
        """
        Generates a color palette for the given categories. This can be used in combination with `node_or_edge_config` to generate
        the dictionary expected by the `colors` argument with colors that cover the whole 0..1 hue range.

        Args:
            cats (iterable): An iterable of categories for which the color palette is to be generated.
            hue (float, optional): The base hue that is added to all the colors in the color palette. It must be in range, 0..1, all the
                resulting hue values will be kept modulo 1.0. Default is 0 (no shift).
            sat (float, optional): The saturation level for the colors. Default is 0.7. Range is 0..1.
            light (float, optional): The lightness level for the colors. Default is 0.5. Range is 0..1.

        Returns:
            (dict): A dictionary where keys are categories and values are hex color codes.
        """
        cats = set(cats)
        cols = {}
        N = len(cats)
        for i, cat in enumerate(cats):
            h = (i/N + hue) % 1.0
            s = sat
            l = light

            r, g, b = self._hsl_to_rgb(h, s, l)

            hex_color = '#%02x%02X%02x' % (r, g, b)
            cols[cat] = hex_color

        return cols


    def node_or_edge_config(self, text_is = None, color_is = None, colors = None, size_is = None, size_range = None, size_scale = 'linear'):
        """
        Create a `node_config` or `edge_config` configuration dictionary for `show()` in an understandable way.

        Args:
            text_is (str): The node/edge attribute to be displayed as text. Use the string `ìd` to draw the node id (regardless of the
                column having another name) or any valid node attribute name.
            color_is (str): A categorical node/edge attribute that can be represented as a color. This will also enable a legend interface
                where categories can be individually shown or hidden.
            colors (dict): The colors for each category defined as a dictionary. The keys are possible outcomes of category.
                The values are html RGB strings. E.g., .draw(category = 'size', colors = {'big' : '#c0a080', 'small' : '#a0c080'})
                where 'big' and 'small' are possible values of the category 'size'.
            size_is (str): The node attribute to be displayed as the size of the nodes. Use the string `id` to set the node id (regardless
                of the column having another name) or any valid node attribute name. See the options in the Moebius configuration menu to
                set minimum, maximum sizes, linear or logarithmic scale, etc.
            size_range (List of two numbers): Combined with edge_label, this parameter controls the values in the variable that
                correspond to the minimum and maximum displayed sizes. The values below or equal the first value will be displayed with the
                base radius (that depends on the zoom) and the values above or equal to the second value will be shown with the maximum
                radius.
            size_scale ('linear', 'power', 'sqrt' or 'log'): Combined with edge_label, the scale used to convert the value in the variable
                to the displayed radius.

        Returns:
            (dict): The node configuration dictionary
        """

        config = {}

        if text_is is not None:
            config['label'] = text_is

        if color_is is not None:
            config['color'] = color_is

        if colors is not None:
            config['color_palette'] = colors
        else:
            config['color_palette'] = {}

        if size_is is None:
            config['size_thresholds'] = []
        else:
            config['size'] = size_is

            if size_range is None:
                config['size_thresholds'] = []
            else:
                assert type(size_range) == list and len(size_range) == 2
                config['size_thresholds'] = size_range

            if size_scale != 'linear':
                assert size_scale in {'power', 'sqrt', 'log'}

            config['scale'] = size_scale

        return config


    def _get_adjacent_nodes_moebius(self, node_id, limit = 20, depth = 1):
        """
        This is the callback function used to interact with the Moebius JS library. Each time you deploy a node in the graph, this
        function is called to get the adjacent nodes and edges.

        The expected return is a JSON string with the following format:

        json = {
            'nodes': [
                {'id' : ..., 'count' : ..., '_int_id' : ...},
                {'id' : ..., 'count' : ..., '_int_id' : ...}
            ],
            'links': [
                {'source' : ..., 'target' : ..., '_int_id' : ...},
                {'source' : ..., 'target' : ..., '_int_id' : ...}
            ]
        },
        where 'count' is the degree of the given node, and '_int_id' is a
        unique integer identifier for the given node or edge.

        Args:
            node_id (str): The id of the node to get the adjacent nodes.
            limit (int): The maximum number of nodes to be returned.
            depth (int): The depth of the graph to be returned.

        Returns:
            (str): A JSON string with the adjacent nodes and edges.
        """

        if self.use_spark:
            nodes_df, edges_df = self._get_one_level_subgraph_graphframes(node_id)
            N = nodes_df.count()

        else:
            nodes_df, edges_df = self._get_one_level_subgraph_networkx(node_id)
            N = len(nodes_df)

        d = 1
        expanded = set([node_id])

        while N < limit and d < depth:
            if self.use_spark:
                next = set(nodes_df.select('id').rdd.flatMap(lambda x: x).collect())
                for id in next:
                    if id not in expanded:
                        expanded.add(id)

                        next_nodes, next_edges = self._get_one_level_subgraph_graphframes(id)
                        nodes_df = nodes_df.union(next_nodes).distinct()
                        edges_df = edges_df.union(next_edges).distinct()

                        N = nodes_df.count()
                        if N >= limit:
                            break
            else:
                next = set(nodes_df.id)
                for id in next:
                    if id not in expanded:
                        expanded.add(id)

                        next_nodes, next_edges = self._get_one_level_subgraph_networkx(id)
                        nodes_df = pd.concat([nodes_df, next_nodes]).drop_duplicates().reset_index(drop = True)
                        edges_df = pd.concat([edges_df, next_edges]).drop_duplicates().reset_index(drop = True)

                        N = len(nodes_df)
                        if N >= limit:
                            break

            d += 1

        if self.use_spark:
            json_final = {
                'nodes': json.loads(nodes_df.toPandas().to_json(orient = 'records', force_ascii = False)),
                'links': json.loads(edges_df.toPandas().to_json(orient = 'records', force_ascii = False))
            }

        else:
            json_final = {
                'nodes': json.loads(nodes_df.to_json(orient = 'records', force_ascii = False)),
                'links': json.loads(edges_df.to_json(orient = 'records', force_ascii = False))
            }

        return json.dumps(json_final, ensure_ascii = False)


    def _get_one_level_subgraph_graphframes(self, node_id, _testing=False):
        """
        Get the one-level subgraph for a given node ID using GraphFrames.

        Args:
            node_id (str): The ID of the node for which to get the one-level subgraph.

        Returns:
            (tuple): A tuple containing two Spark DataFrames:
                - nodes_df: DataFrame with columns 'id', 'count', '_int_id', and any other node attributes.
                - edges_df: DataFrame with columns 'source', 'target', '_int_id' the edges connecting the nodes in the subgraph.
        """

        sql      = SparkInterface().pyspark.sql
        spark    = SparkInterface().spark
        F        = sql.functions
        LongType = sql.types.LongType

        graph = self.G.graphframe

        edges_df = graph.edges.filter((graph.edges.src == node_id) | (graph.edges.dst == node_id))

        N = len(self._int_id_map)
        int_id_map_broadcast = spark.sparkContext.broadcast(self._int_id_map)

        def edge_int_id(src, dst):
            int_id_map = int_id_map_broadcast.value
            return int_id_map[src] + N*(int_id_map[dst] + 1)

        edge_int_id_udf = F.udf(edge_int_id, LongType())

        edges_df = edges_df.withColumn('_int_id', edge_int_id_udf(F.col('src'), F.col('dst')))
        edges_df = edges_df.withColumnRenamed('src', 'source').withColumnRenamed('dst', 'target')

        order = ['source', 'target', '_int_id']
        for col in edges_df.columns:
            if col not in order:
                order.append(col)
        edges_df = edges_df.select(order)

        node_ids = edges_df.select('source').union(edges_df.select('target')).distinct()
        nodes_df = node_ids.join(graph.vertices, node_ids.source == graph.vertices.id, 'inner').select(graph.vertices['*'])

        degrees  = graph.degrees
        # Edges are duplicated in undirected graphframes -> Fix degree
        if not self.G.is_directed:
            degrees = degrees.withColumn('degree', F.ceil(F.col('degree') / 2))
        nodes_df = nodes_df.join(degrees, on = 'id', how = 'left').withColumnRenamed('degree', 'count')

        def node_int_id(id):
            int_id_map = int_id_map_broadcast.value
            return int_id_map[id]

        if _testing:
            key, val = next(iter(self._int_id_map.items()))
            assert node_int_id(key) == val
            assert edge_int_id(key, key) == N*val + N + val

        node_int_id_udf = F.udf(node_int_id, LongType())

        nodes_df = nodes_df.withColumn('_int_id', node_int_id_udf(F.col('id')))

        order = ['id', 'count', '_int_id']
        for col in nodes_df.columns:
            if col not in order:
                order.append(col)
        nodes_df = nodes_df.select(order)

        return nodes_df, edges_df


    def _get_one_level_subgraph_networkx(self, node_id):
        """
        Get the one-level subgraph for a given node ID using Networkx.

        Args:
            node_id (str): The ID of the node for which to get the one-level subgraph.

        Returns:
            (tuple): A tuple containing two Pandas DataFrames:
                - nodes_df: DataFrame with columns 'id', 'count', '_int_id', and any other node attributes.
                - edges_df: DataFrame with columns 'source', 'target', '_int_id' the edges connecting the nodes in the subgraph.
        """

        graph = self.G.networkx

        if self.G.is_directed:
            neighbors = set(graph.neighbors(node_id)) | set(graph.predecessors(node_id))
        else:
            # Undirected graphs are symmetric: predecessors == successors (or neighbors)
            neighbors = set(graph.neighbors(node_id))

        neighbors.add(node_id)

        subgraph = graph.subgraph(neighbors)

        # Create nodes DataFrame
        nodes_data = []
        for node in subgraph.nodes(data = True):
            node_id = node[0]
            attributes = node[1].copy()
            attributes['id'] = node_id
            attributes['count'] = graph.degree[node_id]
            attributes['_int_id'] = self._int_id_map[node_id]
            nodes_data.append(attributes)
        nodes_df = pd.DataFrame(nodes_data)

        order = ['id', 'count', '_int_id']
        for col in nodes_df.columns:
            if col not in order:
                order.append(col)
        nodes_df = nodes_df[order]

        # Create edges DataFrame
        edges_data = []
        N = len(self._int_id_map)
        for edge in subgraph.edges(data = True):
            src, dst, attributes = edge
            attributes = attributes.copy()
            attributes['source'] = src
            attributes['target'] = dst
            attributes['_int_id'] = self._int_id_map[src] + N*(self._int_id_map[dst] + 1)
            edges_data.append(attributes)
        edges_df = pd.DataFrame(edges_data)

        order = ['source', 'target', '_int_id']
        for col in edges_df.columns:
            if col not in order:
                order.append(col)
        edges_df = edges_df[order]

        return nodes_df, edges_df


    def _pd_to_json_format(self, df):
        """
        A utility to produced the flavor of JSON expected by the JS library from a pandas DataFrame.

        Usage:
            ```python
            # This simple snippet to convert the whole graph to the expected JSON format.

            edges = self._pd_to_json_format(self.G.edges_as_pandas().rename({'src' : 'source', 'dst' : 'target'}, axis = 1))
            nodes = self._pd_to_json_format(self.G.nodes_as_pandas())

            json_final = {'nodes' : nodes, 'links' : edges}

            return json.dumps(json_final, ensure_ascii = False)
            ```

        Args:
            df (pd.DataFrame): The DataFrame to be converted to JSON.
        """

        df = df.reset_index(drop = True)

        df_json = df.to_json(orient = 'records', force_ascii = True, date_format = 'iso')

        return json.loads(df_json)


    def _hsl_to_rgb(self, h, s, l):
        """
        Convert HSL (Hue, Saturation, Lightness) color space to RGB (Red, Green, Blue) color space.

        Parameters:
        h (float): Hue value, should be between 0 and 1.
        s (float): Saturation value, should be between 0 and 1.
        l (float): Lightness value, should be between 0 and 1.

        Returns:
        tuple: A tuple containing the RGB values (r, g, b), each ranging from 0 to 255.
        """
        def hue_to_rgb(p, q, t):
            if t < 0: t += 1
            if t > 1: t -= 1
            if t < 1/6: return p + (q - p)*6*t
            if t < 1/2: return q
            if t < 2/3: return p + (q - p)*(2/3 - t)*6
            return p

        q = l + s - l*s if l < 0.5 else l + s - l*s
        p = 2*l - q
        r = hue_to_rgb(p, q, h + 1/3)
        g = hue_to_rgb(p, q, h)
        b = hue_to_rgb(p, q, h - 1/3)

        return int(255*r), int(255*g), int(255*b)


class Moebius(MoebiusAnywidget):
    """
    Moebius class for visualizing graphs using an anywidget.

    **Important**: This class requires: anywidget, traitlets and IPython.display!! These packages are not mandatory
    for the rest of the library, since you may not be intereseted in visualizing graphs. If you want to use this
    class, you need to install these packages.

    Usage:
        ```python
        from mercury.graph.viz import Moebius

        G = ... # A graph object
        moebius = Moebius(G)
        moebius.show()
        ```

    Args:
        G (Graph):          The graph to be visualized.
    """

    def __init__(self, G):
        if display is None or HTML is None:
            raise ImportError('IPython is not installed')

        self.G = G


    def _repr_mimebundle_(self, include=None, exclude=None):
        """
        Override display() method to control its execution and avoid errors.
        """
        self.show()


    def show(self, initial_id = None, initial_depth = 1, node_config = None, edge_config = None):
        """
        Start the interactive graph visualization using an anywidget.

        Args:
            initial_id (str): The id of the node to start the visualization.
            initial_depth (int): The initial depth of the graph (starting with `initial_id` as 0) to be shown.
            node_config (dict): A node configuration dictionary created by `node_config()`.
            edge_config (dict): An edge configuration dictionary created by `edge_config()`.
        """
        display(MoebiusAnywidget(self.G, initial_id, initial_depth, node_config, edge_config))



import inspect

import pandas as pd
import networkx as nx

from mercury.graph.core.spark_interface import SparkInterface, pyspark_installed, graphframes_installed, dgl_installed


class NodeIterator:
    """
    Iterator class for iterating over the nodes in a `mercury.graph.core.Graph`.

    This is returned by the `nodes` property of the `mercury.graph.core.Graph` class.

    Args:
        graph: The `mercury.graph.core.Graph` object to iterate over.

    Usage:

    ```python
    g = mg.Graph(data)
    for node in g.nodes:
        print(node)
    ```
    """

    def __init__(self, graph):
        self.graph = graph
        self.ix = -1
        if graph._as_networkx is not None:
            self.key_list = list(graph._as_networkx.nodes.keys())
        else:
            self.key_list = [row['id'] for row in graph.graphframe.vertices.select('id').collect()]


    def __iter__(self):
        return self


    def __next__(self):
        self.ix += 1

        if self.ix >= len(self.key_list):
            raise StopIteration

        ky = self.key_list[self.ix]

        if self.graph._as_networkx is not None:
            d = self.graph._as_networkx.nodes[ky].copy()
            d['id'] = ky

            return d

        g = self.graph.graphframe
        return g.vertices.filter(g.vertices.id == ky).first().asDict()


class EdgeIterator:
    """
    Iterator class for iterating over the edges in a `mercury.graph.core.Graph`.

    This is returned by the `edges` property of the `mercury.graph.core.Graph` class.

    Args:
        graph: The `mercury.graph.core.Graph` object to iterate over.

    Usage:

    ```python
    g = mg.Graph(data)
    for edge in g.edges:
        print(edge)
    ```
    """

    def __init__(self, graph):
        self.graph = graph
        self.ix = -1
        if graph._as_networkx is not None:
            self.key_list = list(graph._as_networkx.edges.keys())
        else:
            self.key_list = [(row['src'], row['dst']) for row in graph.graphframe.edges.collect()]


    def __iter__(self):
        return self


    def __next__(self):
        self.ix += 1

        if self.ix >= len(self.key_list):
            raise StopIteration

        ky = self.key_list[self.ix]

        if self.graph._as_networkx is not None:
            d = self.graph._as_networkx.edges[ky].copy()
            d['src'] = ky[0]
            d['dst'] = ky[1]

            return d

        g = self.graph.graphframe
        return g.edges.filter((g.edges.src == ky[0]) & (g.edges.dst == ky[1])).first().asDict()


class Graph:
    """ This is the main class in mercury.graph.

    This class seamlessly abstracts the underlying technology used to represent the graph. You can create a graph passing the following
    objects to the constructor:

    - A pandas DataFrame containing edges (with a keys dictionary to specify the columns and possibly a nodes DataFrame)
    - A pyspark DataFrame containing edges (with a keys dictionary to specify the columns and possibly a nodes DataFrame)
    - A networkx graph
    - A graphframes graph

    Bear in mind that the graph object is immutable. This means that you can't modify the graph object once it has been created. If you
    want to modify it, you have to create a new graph object.

    The graph object provides:

    - Properties to access the graph in different formats (networkx, graphframes, dgl)
    - Properties with metrics and summary information that are calculated on demand and technology independent.
    - It is inherited by other graph classes in mercury-graph providing ML algorithms such as graph embedding, visualization, etc.

    Using this class from the other classes in mercury-graph:

    The other classes in mercury-graph define models or functionalities that are based on graphs. They use a Scikit-learn-like API to
    interact with the graph object. This means that the graph object is passed to the class constructor and the class follow the
    Scikit-learn conventions. It is recommended to follow the same conventions when creating your own classes to work with mercury-graph.

    The conventions can be found here:

    - [Scikit API](https://scikit-learn.org/stable/developers/develop.html#apis-of-scikit-learn-objects)
    - [On scikit conventions](https://scikit-learn.org/stable/glossary.html)

    Args:
        data (pd.DataFrame, nx.Graph or pyspark.sql.DataFrame): The data to create the graph from. 
            It can be a pandas DataFrame, a networkx Graph, a pyspark DataFrame, or a Graphframe. 
            In case it already contains a graph (networkx or graphframes), the keys and nodes arguments are ignored.
        keys (dict): A dictionary with keys to specify the columns in the data DataFrame. The keys are:

            - 'src': The name of the column with the source node.
            - 'dst': The name of the column with the destination node.
            - 'id': The name of the column with the node id.
            - 'weight': The name of the column with the edge weight.
            - 'directed': A boolean to specify if the graph is directed. (Only for pyspark DataFrames)

            When the keys argument is not provided or the key is missing, the default values are:

            - 'src': 'src'
            - 'dst': 'dst'
            - 'id': 'id'
            - 'weight': 'weight'
            - 'directed': True

        nodes (pd.DataFrame): A pandas DataFrame or a pyspark DataFrame with the nodes data. (Only when `data` is pandas or pyspark DataFrame and with the
            same type as `data`) If not given, the nodes are inferred from the edges DataFrame.
    """
    def __init__(self, data = None, keys = None, nodes = None):
        self._as_networkx = None
        self._as_graphframe = None
        self._as_dgl = None
        self._degree = None
        self._in_degree = None
        self._out_degree = None
        self._closeness_centrality = None
        self._betweenness_centrality = None
        self._pagerank = None
        self._connected_components = None
        self._nodes_colnames = None
        self._edges_colnames = None

        self._number_of_nodes = 0
        self._number_of_edges = 0
        self._node_ix = 0
        self._is_directed = False
        self._is_weighted = False

        self._init_values = {k: v for k, v in locals().items() if k in inspect.signature(self.__init__).parameters}

        if type(data) == pd.core.frame.DataFrame:
            self._from_pandas(data, nodes, keys)
            return

        if isinstance(data, nx.Graph):      # This is the most general case, including: ...Graph, ...DiGraph and ...MultiGraph
            self._from_networkx(data)
            return

        spark_int = SparkInterface()

        if pyspark_installed and graphframes_installed:
            if type(data) in spark_int.type_spark_dataframe:
                self._from_dataframe(data, nodes, keys)
                return

            if type(data) == spark_int.type_graphframe:
                self._from_graphframes(data)
                return

        raise ValueError('Invalid input data. (Expected: pandas DataFrame, a networkx Graph, a pyspark DataFrame, a graphframes Graph.)')


    def __str__(self):
        txt = []
        txt.append('mercury.graph.core.Graph with %d nodes and %d edges.' % (self._number_of_nodes, self._number_of_edges))
        txt.append('  is_directed:    %s' % self._is_directed)
        txt.append('  is_weighted:    %s' % self._is_weighted)
        txt.append('  has_networkx:   %s' % (self._as_networkx is not None))
        txt.append('  has_graphframe: %s' % (self._as_graphframe is not None))
        txt.append('  has_dgl:        %s' % (self._as_dgl is not None))

        return '\n'.join(txt)


    def __repr__(self):
        return 'Graph(%s)' % ', '.join('%s = %s' % (k, v) for k, v in self._init_values.items())


    @property
    def nodes(self):
        """
        Returns an iterator over all the nodes in the graph.

        Returns:
            (NodeIterator): An iterator that yields each node in the graph.
        """
        return NodeIterator(self)


    @property
    def edges(self):
        """
        Returns an iterator over the edges in the graph.

        Returns:
            (EdgeIterator): An iterator object that allows iterating over the edges in the graph.
        """
        return EdgeIterator(self)


    @property
    def networkx(self):
        """
        Returns the graph representation as a NetworkX graph.

        If the graph has not been converted to NetworkX format yet, it will be converted and cached for future use.

        Returns:
            (networkx.Graph): The graph representation as a NetworkX graph.
        """
        if self._as_networkx is None:
            self._as_networkx = self._to_networkx()

        return self._as_networkx


    @property
    def graphframe(self):
        """
        Returns the graph as a GraphFrame.

        If the graph has not been converted to a GraphFrame yet, it will be converted and cached for future use.

        Returns:
            (GraphFrame): The graph represented as a GraphFrame.
        """
        if self._as_graphframe is None:
            self._as_graphframe = self._to_graphframe()

        return self._as_graphframe


    @property
    def dgl(self):
        """
        Returns the graph as a DGL graph.

        If the graph has not been converted to a DGL graph yet, it will be converted and cached for future use.

        Returns:
            (dgl.DGLGraph): The graph represented as a DGL graph.
        """
        if self._as_dgl is None:
            self._as_dgl = self._to_dgl()

        return self._as_dgl


    @property
    def degree(self):
        """
        Returns the degree of each node in the graph as a Python dictionary.
        """
        if self._degree is None:
            self._degree = self._calculate_degree()
        return self._degree


    @property
    def in_degree(self):
        """
        Returns the in-degree of each node in the graph as a Python dictionary.
        """
        if self._in_degree is None:
            self._in_degree = self._calculate_in_degree()
        return self._in_degree


    @property
    def out_degree(self):
        """
        Returns the out-degree of each node in the graph as a Python dictionary.
        """
        if self._out_degree is None:
            self._out_degree = self._calculate_out_degree()
        return self._out_degree


    @property
    def closeness_centrality(self):
        """
        Returns the closeness centrality of each node in the graph as a Python dictionary.
        """
        if self._closeness_centrality is None:
            self._closeness_centrality = self._calculate_closeness_centrality()
        return self._closeness_centrality


    @property
    def betweenness_centrality(self):
        """
        Returns the betweenness centrality of each node in the graph as a Python dictionary.
        """
        if self._betweenness_centrality is None:
            self._betweenness_centrality = self._calculate_betweenness_centrality()
        return self._betweenness_centrality


    @property
    def pagerank(self):
        """
        Returns the PageRank of each node in the graph as a Python dictionary.
        """
        if self._pagerank is None:
            self._pagerank = self._calculate_pagerank()
        return self._pagerank


    @property
    def connected_components(self):
        """
        Returns the connected components of each node in the graph as a Python dictionary.
        """
        if self._connected_components is None:
            self._connected_components = self._calculate_connected_components()
        return self._connected_components


    @property
    def nodes_colnames(self):
        """
        Returns the column names of the nodes DataFrame.
        """
        if self._nodes_colnames is None:
            self._nodes_colnames = self._calculate_nodes_colnames()
        return self._nodes_colnames


    @property
    def edges_colnames(self):
        """
        Returns the column names of the edges DataFrame.
        """
        if self._edges_colnames is None:
            self._edges_colnames = self._calculate_edges_colnames()
        return self._edges_colnames


    @property
    def number_of_nodes(self):
        """
        Returns the number of nodes in the graph.

        Returns:
            (int): The number of nodes in the graph.
        """
        return self._number_of_nodes


    @property
    def number_of_edges(self):
        """
        Returns the number of edges in the graph.

        Returns:
            (int): The number of edges in the graph.
        """
        return self._number_of_edges


    @property
    def is_directed(self):
        """
        Returns True if the graph is directed, False otherwise.

        Note: 
            Graphs created using graphframes are always directed. The way around it is to add the reverse edges to the graph.
            This can be done by creating the Graph with pyspark DataFrame() and defining a key 'directed' set as False in the `dict`
            argument. Otherwise, the graph will be considered directed even if these reversed edges have been created by other means
            this class cannot be aware of.
        """
        return self._is_directed


    @property
    def is_weighted(self):
        """
        Returns True if the graph is weighted, False otherwise.

        A graph is considered weight if it has a column named 'weight' in the edges DataFrame or the column has a different name and that
        name is passed in the `dict` argument as the 'weight' key.
        """
        return self._is_weighted


    def nodes_as_pandas(self):
        """
        Returns the nodes as a pandas DataFrame.

        If the graph is represented as a networkx graph, the nodes are extracted from it. Otherwise, the graphframes graph will be used.
        This dataset may differ from possible pandas DataFrame passed to the constructor in the column names and order. The column used
        as the node id is always named 'id'.
        """
        if self._as_networkx is not None:
            nodes_data = self._as_networkx.nodes(data = True)
            nodes_df   = pd.DataFrame([(node, attr) for node, attr in nodes_data], columns = ['id', 'attributes'])

            attrs_df = pd.json_normalize(nodes_df['attributes'])

            return pd.concat([nodes_df.drop('attributes', axis = 1), attrs_df], axis = 1)

        return self.graphframe.vertices.toPandas()


    def edges_as_pandas(self):
        """
        Returns the edges as a pandas DataFrame.

        If the graph is represented as a networkx graph, the edges are extracted from it. Otherwise, the graphframes graph will be used.
        This dataset may differ from possible pandas DataFrame passed to the constructor in the column names and order. The columns used
        as the source and destination nodes are always named 'src' and 'dst', respectively.
        """
        if self._as_networkx is not None:
            edges_data = self._as_networkx.edges(data = True)
            edges_df   = pd.DataFrame([(src, dst, attr) for src, dst, attr in edges_data], columns = ['src', 'dst', 'attributes'])

            attrs_df   = pd.json_normalize(edges_df['attributes'])

            return pd.concat([edges_df.drop('attributes', axis = 1), attrs_df], axis = 1)

        return self.graphframe.edges.toPandas()


    def nodes_as_dataframe(self):
        """
        Returns the nodes as a pyspark DataFrame.

        If the graph is represented as a graphframes graph, the nodes are extracted from it. Otherwise, the nodes are converted from the
        pandas DataFrame representation. The column used as the node id is always named 'id', regardless of the original column name passed
        to the constructor.
        """
        if self._as_graphframe is not None:
            return self._as_graphframe.vertices

        return SparkInterface().spark.createDataFrame(self.nodes_as_pandas())


    def edges_as_dataframe(self):
        """
        Returns the edges as a pyspark DataFrame.

        If the graph is represented as a graphframes graph, the edges are extracted from it. Otherwise, the edges are converted from the
        pandas DataFrame representation. The columns used as the source and destination nodes are always named 'src' and 'dst',
        respectively, regardless of the original column names passed to the constructor.
        """
        if self._as_graphframe is not None:
            return self._as_graphframe.edges

        return SparkInterface().spark.createDataFrame(self.edges_as_pandas())


    def _from_pandas(self, edges, nodes, keys):
        """ This internal method extends the constructor to accept a pandas DataFrame as input.

        It takes the constructor arguments and does not return anything. It sets the internal state of the object.
        """
        if keys is None:
            src = 'src'
            dst = 'dst'
            id  = 'id'
            weight = 'weight'
            directed = True
        else:
            src = keys.get('src', 'src')
            dst = keys.get('dst', 'dst')
            id  = keys.get('id', 'id')
            weight = keys.get('weight', 'weight')
            directed = keys.get('directed', True)

        if directed:
            g = nx.DiGraph()
        else:
            g = nx.Graph()

        if weight in edges.columns:
            edges = edges.rename(columns = {weight: 'weight'})

        for _, row in edges.iterrows():
            attr = row.drop([src, dst]).to_dict()
            g.add_edge(row[src], row[dst], **attr)

        if nodes is not None:
            for _, row in nodes.iterrows():
                attr = row.drop([id]).to_dict()
                g.add_node(row[id], **attr)

        self._from_networkx(g)


    def _from_dataframe(self, edges, nodes, keys):
        """ This internal method extends the constructor to accept a pyspark DataFrame as input.

        It takes the constructor arguments and does not return anything. It sets the internal state of the object.
        """
        if keys is None:
            src = 'src'
            dst = 'dst'
            id  = 'id'
            weight = 'weight'
            directed = True
        else:
            src = keys.get('src', 'src')
            dst = keys.get('dst', 'dst')
            id  = keys.get('id', 'id')
            weight = keys.get('weight', 'weight')
            directed = keys.get('directed', True)

        edges = edges.withColumnRenamed(src, 'src').withColumnRenamed(dst, 'dst')

        if weight in edges.columns:
            edges = edges.withColumnRenamed(weight, 'weight')

        if nodes is not None:
            nodes = nodes.withColumnRenamed(id, 'id').dropDuplicates(['id'])
        else:
            src_nodes = edges.select('src').distinct().withColumnRenamed('src', 'id')
            dst_nodes = edges.select('dst').distinct().withColumnRenamed('dst', 'id')
            nodes = src_nodes.union(dst_nodes).distinct()

        g = SparkInterface().graphframes.GraphFrame(nodes, edges)

        if not directed:
            edges = g.edges

            other_columns = [col for col in edges.columns if col not in ('src', 'dst')]
            reverse_edges = edges.select(edges['dst'].alias('src'), edges['src'].alias('dst'), *other_columns)
            all_edges     = edges.union(reverse_edges).distinct()

            g = SparkInterface().graphframes.GraphFrame(nodes, all_edges)

        self._from_graphframes(g, directed)


    def _from_networkx(self, graph):
        """ This internal method extends the constructor to accept a networkx graph as input.

        It takes the constructor arguments and does not return anything. It sets the internal state of the object.
        """
        self._as_networkx = graph
        self._number_of_nodes = len(graph.nodes)
        self._number_of_edges = len(graph.edges)
        self._is_directed = nx.is_directed(graph)
        self._is_weighted = 'weight' in self.edges_colnames


    def _from_graphframes(self, graph, directed = True):
        """ This internal method extends the constructor to accept a graphframes graph as input.

        It takes the constructor arguments and does not return anything. It sets the internal state of the object.
        """
        self._as_graphframe = graph
        self._number_of_nodes = graph.vertices.count()
        self._number_of_edges = graph.edges.count()
        self._is_directed = directed
        self._is_weighted = 'weight' in self.edges_colnames


    def _to_networkx(self):
        """ This internal method handles the logic of a property. It returns the networkx graph that already exists
        or converts it from the graphframes graph if not."""

        if self._is_directed:
            g = nx.DiGraph()
        else:
            g = nx.Graph()

        for _, row in self.edges_as_pandas().iterrows():
            attr = row.drop(['src', 'dst']).to_dict()
            g.add_edge(row['src'], row['dst'], **attr)

        for _, row in self.nodes_as_pandas().iterrows():
            attr = row.drop(['id']).to_dict()
            g.add_node(row['id'], **attr)

        return g


    def _to_graphframe(self):
        """ This internal method handles the logic of a property. It returns the graphframes graph that already exists
        or converts it from the networkx graph if not."""

        nodes = self.nodes_as_dataframe()
        edges = self.edges_as_dataframe()

        return SparkInterface().graphframes.GraphFrame(nodes, edges)


    def _to_dgl(self):
        """ This internal method handles the logic of a property. It returns the dgl graph that already exists
        or converts it from the networkx graph if not."""

        if dgl_installed:
            dgl = SparkInterface().dgl

            edge_attrs = [c for c in self.edges_colnames if c not in ['src', 'dst']]
            if len(edge_attrs) == 0:
                edge_attrs = None

            node_attrs = [c for c in self.nodes_colnames if c not in ['id']]
            if len(node_attrs) == 0:
                node_attrs = None

            self._as_dgl = dgl.from_networkx(self.networkx, edge_attrs = edge_attrs, node_attrs = node_attrs)

        return self._as_dgl


    def _calculate_degree(self):
        """ This internal method handles the logic of a property. It returns the degree of each node in the graph."""

        if self._as_networkx is not None:
            return dict(self._as_networkx.degree())

        return self._fill_node_zeros({row['id']: row['degree'] for row in self.graphframe.degrees.collect()})


    def _calculate_in_degree(self):
        """ This internal method handles the logic of a property. It returns the in-degree of each node in the graph."""

        if self._as_networkx is not None:
            return dict(self._as_networkx.in_degree())

        return self._fill_node_zeros({row['id']: row['inDegree'] for row in self.graphframe.inDegrees.collect()})


    def _calculate_out_degree(self):
        """ This internal method handles the logic of a property. It returns the out-degree of each node in the graph."""

        if self._as_networkx is not None:
            return dict(self._as_networkx.out_degree())

        return self._fill_node_zeros({row['id']: row['outDegree'] for row in self.graphframe.outDegrees.collect()})


    def _fill_node_zeros(self, d):
        """
        This internal method fills the nodes that are not in the dictionary with a zero value. This make the output obtained from
        graphframes consistent with the one from networkx.
        """
        for node in self.nodes:
            if node['id'] not in d:
                d[node['id']] = 0

        return d


    def _calculate_closeness_centrality(self):
        """
        This internal method handles the logic of a property. It returns the closeness centrality of each node in the graph as
        a Python dictionary.
        """
        if self._as_networkx is not None:
            return nx.closeness_centrality(self._as_networkx)

        nodes = [row['id'] for row in self.graphframe.vertices.select('id').collect()]
        paths = self.graphframe.shortestPaths(landmarks = nodes)
        expr  = SparkInterface().pyspark.sql.functions.expr
        sums  = paths.withColumn('sums', expr('aggregate(map_values(distances), 0, (acc, x) -> acc + x)'))

        cc = sums.withColumn('cc', (self.number_of_nodes - 1)/sums['sums']).select('id', 'cc')

        return {row['id']: row['cc'] for row in cc.collect()}


    def _calculate_betweenness_centrality(self):
        """
        This internal method handles the logic of a property. It returns the betweenness centrality of each node in the graph as a Python
        dictionary. NOTE: This method converts the graph to a networkx graph to calculate the betweenness centrality since the algorithm
        is too computationally expensive to use on large graphs.
        """
        return nx.betweenness_centrality(self.networkx)


    def _calculate_pagerank(self):
        """
        This internal method handles the logic of a property. It returns the PageRank of each node in the graph as a Python dictionary.
        """
        if self._as_networkx is not None:
            return nx.pagerank(self._as_networkx)

        pr = self.graphframe.pageRank(resetProbability = 0.15, tol = 0.01).vertices

        return {row['id']: row['pagerank'] for row in pr.collect()}


    def _calculate_connected_components(self):
        """
        This internal method handles the logic of a property. It returns the connected components of each node in the graph as a Python
        dictionary.
        """
        if self._as_networkx is not None:
            if self._is_directed:
                G = self._as_networkx.to_undirected()
            else:
                G = self._as_networkx

            graphs = (G.subgraph(c) for c in nx.connected_components(G))
            cc = dict()
            for i, graph in enumerate(graphs):
                n = graph.number_of_nodes()
                for nid in graph.nodes:
                    cc[nid] = {'cc_id' : i, 'cc_size' : n}

            return cc

        graphs = self.graphframe.connectedComponents(algorithm = 'graphx')
        cc_size = graphs.select('id', 'component').groupBy('component').count()
        cc_all = graphs.select('id', 'component').join(cc_size, 'component', how = 'left_outer')

        cc = dict()
        for row in cc_all.collect():
            cc[row['id']] = {'cc_id' : row['component'], 'cc_size' : row['count']}

        return cc


    def _calculate_nodes_colnames(self):
        """ This internal method returns the column names of the nodes DataFrame. """

        if self._as_networkx is not None:
            l = ['id']
            l.extend(list(self._as_networkx.nodes[list(self._as_networkx.nodes.keys())[0]].keys()))

            return l

        return self.graphframe.vertices.columns


    def _calculate_edges_colnames(self):
        """ This internal method returns the column names of the edges DataFrame. """

        if self._as_networkx is not None:
            l = ['src', 'dst']
            l.extend(list(self._as_networkx.edges[list(self._as_networkx.edges.keys())[0]].keys()))
            return l

        return self.graphframe.edges.columns



from mercury.graph.core import Graph
from mercury.graph.core.base import BaseClass

from mercury.graph.core.spark_interface import pyspark_installed, graphframes_installed

if pyspark_installed:
    import pyspark.sql.functions as f
    from pyspark.sql import Window
    from mercury.graph.embeddings.spark_node2vec import udf_select_element_2

    if graphframes_installed:
        from graphframes import GraphFrame
        from graphframes.lib import AggregateMessages


class SparkRandomWalker(BaseClass):

    def __init__(self, num_epochs=10, batch_size=1, n_sampling_edges=None):
        """
        Class to perform random walks from a specific source_id node within a given Graph

        Args:
            num_epochs (int): Number of epochs. This is the total number of steps the iteration goes through.
            batch_size (int): This forces caching the random walks computed so far and breaks planning each time this number of epochs
                is reached. The default value is a high number to avoid this entering at all. In really large jobs, you may want to
                set this parameter to avoid possible overflows even if it can add some extra time to the process. Note that with a high
                number of epochs and nodes resource requirements for the active part of your random walks can be high. This allows to
                "cache a continue" so to say.
            n_sampling_edges (int): by setting this parameter you can limit at each timestep the number of new paths opened from each node.
                This is useful when the graph contains nodes with very high out-degree, where running the algorithm several epochs is
                not feasible. When using this parameter, the graph will consider only at most `edge_sampling` outgoing edges at each
                epoch for each path. If the last node of the path contains more than `edge_sampling` the selected edges are sampled
                using its weight.
        """
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.n_sampling_edges = n_sampling_edges

    def __str__(self):
        base_str = super().__str__()

        # Check if the object has been fitted (fitting creates `paths` attribute)
        if hasattr(self, "paths_"):
            extra_str = [f"", f"Random walk paths are available in attribute `paths_`."]
            return "\n".join([base_str] + extra_str)
        else:
            return base_str

    def fit(self, G: Graph, source_id):
        """
        Perform random walks from a specific source_id node within a given Graph

        Args:
            G (mercury.graph Graph asset): A `mercury.graph` Graph
            source_id (int/str/list): the source vertex or list for vertices to start the random walks.

        Returns:
            (self): Fitted self (or raises an error)

        Attribute `paths_` contains a Spark Dataframe with a columns `random_walks` containing an array of the elements
        of the path walked and another column with the corresponding weights. The weights represent the probability of
        following that specific path starting from source_id.
        """
        self.paths_ = self._run_rw(G, source_id)

        return self

    def _run_rw(self, G: Graph, source_id):
        self._start_rw(G, source_id)

        for i in range(self.num_epochs):

            aux_vert = self._update_state_with_next_step(i)
            self.gx = GraphFrame(aux_vert, self.gx.edges)

            if (i + 1) % self.batch_size == 0:
                old_aux_vert = aux_vert
                aux_vert = AggregateMessages.getCachedDataFrame(aux_vert)
                old_aux_vert.unpersist()

        paths = (
            self.gx.vertices.select(
                f.col("new_rw_acc_path").alias("random_walks"),
                f.col("new_rw_acc_weight").alias("weights"),
            )
            .filter(f.col("random_walks").isNotNull())
            .persist()
        )

        self.gx.vertices.unpersist()
        self.gx.edges.unpersist()

        return paths

    def _start_rw(self, G: Graph, source_id):
        aux_vert = (
            G.graphframe.vertices.groupBy(f.col("id"))
            .agg(f.collect_list(f.col("id")).alias("tmp_rw_aux_acc_path"))
            .withColumn("in_sample", f.col("id").isin(source_id))
            .withColumn(
                "new_rw_acc_path",
                f.when(f.col("in_sample"), f.col("tmp_rw_aux_acc_path")).otherwise(
                    f.lit(None)
                ),
            )
            .withColumn("new_rw_curr_id", f.col("id"))
            .drop("tmp_rw_aux_acc_path", "in_sample")
        ).filter(f.col("new_rw_acc_path").isNotNull())
        aux_vert = aux_vert.withColumn("new_rw_acc_weight", f.array(f.lit(1.0)))

        w_ind_edge = Window.partitionBy("src")

        df_edges_cumsum = G.graphframe.edges.withColumn(
            "new_rw_total_sum", f.sum("weight").over(w_ind_edge)
        )

        aux_edges = (
            df_edges_cumsum.withColumn(
                "new_rw_norm_sum",
                f.col("weight") / f.col("new_rw_total_sum"),
            ).select(
                f.col("src"),
                f.col("dst"),
                f.col("weight"),
                f.col("new_rw_norm_sum"),
            )
        ).persist()

        self.gx = GraphFrame(aux_vert, aux_edges)

    def _update_state_with_next_step(self, i):
        candidate_vert_col = self.gx.vertices.columns

        candidate_vert = self.gx.vertices.withColumnRenamed(
            "new_rw_curr_id", "new_rw_prev_id"
        )

        if self.n_sampling_edges:
            out_edges = self._sample_edges()
        else:
            out_edges = self.gx.edges

        candidate_vert_with_edges = candidate_vert.join(
            out_edges,
            f.col("new_rw_prev_id") == f.col("src"),
            "left",
        )

        selected_next_step = (
            candidate_vert_with_edges.withColumn(
                "new_rw_acc_path",
                udf_select_element_2(f.col("new_rw_acc_path"), f.col("dst")),
            )
            .withColumn(
                "new_rw_acc_weight",
                udf_select_element_2(
                    f.col("new_rw_acc_weight"),
                    f.col("new_rw_acc_weight")[f.size("new_rw_acc_weight") - f.lit(1)]
                    * f.col("new_rw_norm_sum"),
                ),
            )
            .withColumn("new_rw_curr_id", f.col("new_rw_acc_path").getItem(i + 1))
            .select(candidate_vert_col)
        )

        return selected_next_step

    def _sample_edges(self):

        sampled_edges = self.gx.edges

        # For each edge, Generate random number and multiply by the edge weight
        sampled_edges = sampled_edges.withColumn("rnd_e", f.rand())
        sampled_edges = sampled_edges.withColumn(
            "weighted_rnd", f.col("rnd_e") * f.col("new_rw_norm_sum")
        )

        # Sort by weighted_rnd and select the top edge_sampling edges
        win_steps = Window.partitionBy("src").orderBy(f.col("weighted_rnd").desc())
        sampled_edges = sampled_edges.withColumn(
            "row_number", f.row_number().over(win_steps)
        )
        sampled_edges = sampled_edges.filter(
            f.col("row_number") <= self.n_sampling_edges
        )
        return sampled_edges



from mercury.graph.core.graph import Graph
from mercury.graph.core.spark_interface import SparkInterface
from mercury.graph.core.spark_interface import pyspark_installed, graphframes_installed, default_spark_config, dgl_installed
from mercury.graph.core._njit import njit, graph_i4, numba_installed



from mercury.graph.ml.spectral import SpectralClustering
from mercury.graph.ml.louvain import LouvainCommunities
from mercury.graph.ml.transition import Transition
from mercury.graph.ml.spark_randomwalker import SparkRandomWalker
from mercury.graph.ml.spark_spreadactivation import SparkSpreadingActivation
from mercury.graph.ml.graph_features import GraphFeatures



from mercury.graph.core import Graph
from mercury.graph.core.base import BaseClass
from mercury.graph.core.spark_interface import pyspark_installed

from pandas import DataFrame
from networkx import normalized_laplacian_matrix
from networkx.algorithms.community import modularity as nx_modularity

from sklearn.cluster import KMeans
from numpy.linalg import eigh
from numpy import asarray
import numpy as np

if pyspark_installed:
    from pyspark.sql import functions as F
    from pyspark.ml.clustering import PowerIterationClustering

class SpectralClustering(BaseClass):
    """
    Implementation of the spectral clustering algorithm which detect communities inside a graph.

    Contributed by Gibran Gabriel Otazo Sanchez.

    Args:
        n_clusters (int): The number of clusters that you want to detect.
        random_state (int): Seed for reproducibility
        mode (str): Calculation mode. Pass 'networkx' for using pandas + networkx or
                    'spark' for spark + graphframes
        max_iterations (int): Max iterations parameter (only used if mode==spark)
    """

    def __init__(
        self, n_clusters=2, mode="networkx", max_iterations=10, random_state=0
    ):
        self.n_clusters = n_clusters
        self.mode = mode
        self.max_iterations = max_iterations
        self.random_state = random_state

        if self.mode not in ("networkx", "spark"):
            raise ValueError("Error: Mode must be either 'networkx' or 'spark'")

    def __str__(self):
        base_str = super().__str__()

        # Check if the object has been fitted (fitting creates the `labels_` attribute)
        if hasattr(self, "labels_"):
            extra_str = [
                f"",
                f"Cluster assignments are available in attribute `labels_`",
                f"Modularity: {self.modularity_}",
            ]
            return "\n".join([base_str] + extra_str)
        else:
            return base_str

    def fit(self, graph: Graph):
        """
        Find the optimal clusters of a given graph. The function returns nothing, but saves the clusters and
        the modularity in the object self.

        Args:
            graph (Graph): A mercury graph structure.

        Returns:
            (self): Fitted self (or raises an error)

        """
        if self.mode == "networkx":
            self._fit_networkx(graph)
        else:
            self._fit_spark(graph)

        return self

    def _fit_networkx(self, graph: Graph):
        """
        Spectral clustering but using networkx (local mode implementation)

        Args:
            graph (Graph): A mercury graph structure.

        Returns:
            (self): Fitted self (or raises an error)
        """
        gnx = graph.networkx.to_undirected()

        L = normalized_laplacian_matrix(gnx).todense()

        if not np.allclose(L, L.T):
            raise ValueError("Normalized Laplacian matrix of the undirected graph should be symmetric")

        w, v = eigh(L)

        U = v[:, : self.n_clusters]
        U = asarray(U)

        kmeans = KMeans(
            n_clusters=self.n_clusters, random_state=self.random_state, n_init="auto"
        ).fit(U)

        self.labels_ = DataFrame({"node_id": gnx.nodes(), "cluster": kmeans.labels_})

        cluster_nodes = self.labels_.groupby("cluster")["node_id"].apply(list)
        self.modularity_ = nx_modularity(gnx, cluster_nodes)

    def _fit_spark(self, graph: Graph):
        """
        Spectral clustering but using pyspark

        Args:
            graph (Graph): A mercury graph structure.

        Returns:
            (self): Fitted self (or raises an error)
        """

        graph_frames_graph = graph.graphframe

        pic = PowerIterationClustering(k=self.n_clusters, weightCol="weight")
        pic.setMaxIter(self.max_iterations)

        # Node ids can be strings, with this we ensure IDs are always converted to
        # integers (needed by PowerIterationClustering)
        vertices_mapping = graph_frames_graph.vertices.withColumn(
            "idx", F.monotonically_increasing_id()
        )

        mapped_node_ids = (
            graph_frames_graph.edges.join(
                vertices_mapping, graph_frames_graph.edges.src == vertices_mapping.id
            )
            .withColumnRenamed("idx", "src_mapped")
            .drop("id", "src")
        )

        mapped_node_ids = (
            mapped_node_ids.join(
                vertices_mapping, mapped_node_ids.dst == vertices_mapping.id
            )
            .withColumnRenamed("idx", "dst_mapped")
            .drop("id", "dst")
            .withColumnRenamed("src_mapped", "src")
            .withColumnRenamed("dst_mapped", "dst")
        )
        assignments = pic.assignClusters(mapped_node_ids)

        self.labels_ = (
            vertices_mapping.join(assignments, vertices_mapping.idx == assignments.id)
            .drop(assignments.id)
            .selectExpr(["id as node_id", "cluster"])
        )

        self.modularity_ = self._spark_modularity(
            graph_frames_graph.edges, graph_frames_graph.degrees
        )

    def _spark_modularity(self, edges, degrees, resolution=1):
        """Computes modularity using the same approximation as networkx:
        https://networkx.org/documentation/stable/reference/algorithms/generated/networkx.algorithms.community.quality.modularity.html
        """

        edge_nb = edges.count()
        q = []

        for i in range(self.n_clusters):
            nids = self.labels_[self.labels_.cluster == i]
            nodeids = [row["node_id"] for row in nids.select("node_id").collect()]

            l_c = edges.filter(
                edges.src.isin(nodeids) & edges.dst.isin(nodeids)
            ).count()

            k_c = (
                nids.join(degrees.withColumnRenamed("id", "node_id"), on="node_id")
                .select(F.sum("degree"))
                .collect()[0][0]
            )

            qi = (l_c / edge_nb) - resolution * (k_c / (2 * edge_nb)) ** 2
            q.append(qi)

        return np.sum(q)



import importlib.util
import tempfile

pyspark_installed = False
graphframes_installed = False
dgl_installed = False

if importlib.util.find_spec("pyspark") is not None:
    pyspark_installed = True

    import pyspark

    if importlib.util.find_spec("graphframes") is not None:
        graphframes_installed = True

        import graphframes as gf

if importlib.util.find_spec("dgl") is not None:
    dgl_installed = True

    import dgl


# Define the Spark configuration options by default
default_spark_config = {
    "appName": "LocalSparkApp",
    "master": "local[*]",
    # Add other configurations as needed
    # 'spark.executor.memory': '2g',
    # 'spark.driver.memory': '1g',
}


class SparkInterface:
    """
    A class that provides an interface for interacting with Apache Spark, graphframes and dgl.

    Attributes:
        _spark_session (pyspark.sql.SparkSession): The shared Spark session.
        _graphframes (module): The shared graphframes namespace.

    Methods:
        __init__(self, config=None): Initializes the SparkInterface object.
        _create_spark_session(config): Creates a Spark session.
        spark: Property that returns the shared Spark session.
        pyspark: Property that returns the pyspark namespace.
        graphframes: Property that returns the shared graphframes namespace.
        dgl: Property that returns the shared dgl namespace.
        read_csv(path, **kwargs): Reads a CSV file into a DataFrame.
        read_parquet(path, **kwargs): Reads a Parquet file into a DataFrame.
        read_json(path, **kwargs): Reads a JSON file into a DataFrame.
        read_text(path, **kwargs): Reads a text file into a DataFrame.
        read(path, format, **kwargs): Reads a file into a DataFrame.
        sql(query): Executes a SQL query.
        udf(f, returnType): Registers a user-defined function (UDF).
        stop(): Stops the Spark session.

    Args:
        config (dict, optional): A dictionary of Spark configuration options.
            If not provided, the configuration in the global variable `default_spark_config` will be used.
    """

    _spark_session = None  # Class variable to hold the shared Spark session
    _graphframes = None  # Class variable to hold the shared graphframes namespace
    _dgl = None  # Class variable to hold the shared dgl namespace

    def __init__(self, config=None, session=None):
        if SparkInterface._spark_session is None:
            if session is not None:
                SparkInterface._spark_session = session
            else:
                SparkInterface._spark_session = self._create_spark_session(config)
                # Set checkpoint directory
                checkpoint_dir = tempfile.mkdtemp()
                SparkInterface._spark_session.sparkContext.setCheckpointDir(checkpoint_dir)

        if SparkInterface._graphframes is None and graphframes_installed:
            SparkInterface._graphframes = gf

        if SparkInterface._dgl is None and dgl_installed:
            SparkInterface._dgl = dgl

    @staticmethod
    def _create_spark_session(config):
        if config is None:
            config = default_spark_config

        spark_builder = pyspark.sql.SparkSession.builder

        for key, value in config.items():
            spark_builder = spark_builder.config(key, value)

        return spark_builder.getOrCreate()

    @property
    def spark(self):
        return SparkInterface._spark_session

    @property
    def pyspark(self):
        return pyspark

    @property
    def graphframes(self):
        return SparkInterface._graphframes

    @property
    def dgl(self):
        return SparkInterface._dgl

    @property
    def type_spark_dataframe(self):
        ret = [pyspark.sql.dataframe.DataFrame]
        try:
            ret.append(pyspark.sql.connect.DataFrame)
        except AttributeError:
            pass

        return ret

    @property
    def type_graphframe(self):
        return gf.graphframe.GraphFrame

    def read_csv(self, path, **kwargs):
        return self.spark.read.csv(path, **kwargs)

    def read_parquet(self, path, **kwargs):
        return self.spark.read.parquet(path, **kwargs)

    def read_json(self, path, **kwargs):
        return self.spark.read.json(path, **kwargs)

    def read_text(self, path, **kwargs):
        return self.spark.read.text(path, **kwargs)

    def read(self, path, format, **kwargs):
        return self.spark.read.format(format).load(path, **kwargs)

    def sql(self, query):
        return self.spark.sql(query)

    def udf(self, f, returnType):
        return self.spark.udf.register(f.__name__, f, returnType)



import bz2
import pickle

import numpy as np
import pandas as pd
import networkx as nx

from mercury.graph.core import Graph, njit, graph_i4
from mercury.graph.core.base import BaseClass
from mercury.graph.embeddings import Embeddings


@njit
def _random_node_weighted(r_sum, TotW):
    r = TotW * np.random.random() - 1e-8
    i = 0
    while r > r_sum[i]:
        r -= r_sum[i]
        i += 1

    return i


@njit
def _random_walks(r_ini, r_len, r_sum, r_col, r_wgt, TotW, n_jmp, max_jpe):
    N = len(r_ini)

    convrge = np.zeros((n_jmp, 2), dtype=graph_i4)
    diverge = np.zeros((n_jmp, 2), dtype=graph_i4)

    ori = _random_node_weighted(r_sum, TotW)
    njm = 0

    for t in range(n_jmp):

        while r_len[ori] == 0 or njm >= max_jpe:
            ori = _random_node_weighted(r_sum, TotW)
            njm = 0

        njm += 1

        rnd = r_sum[ori] * np.random.random()
        rst = r_ini[ori]
        for ic in range(r_len[ori]):
            nxt = r_col[rst + ic]
            wei = r_wgt[rst + ic]

            if wei >= rnd:
                break

            rnd -= wei

        convrge[t, 0] = ori
        convrge[t, 1] = nxt

        oth = int(N * np.random.random())
        while oth == ori or oth == nxt:
            oth = int(N * np.random.random())

        diverge[t, 0] = ori
        diverge[t, 1] = oth

        ori = nxt

    return (convrge, diverge)


class GraphEmbedding(BaseClass):
    """
    Create an embedding mapping the nodes of a graph.

    Includes contributions by David Muelas Recuenco.

    Args:
        dimension (int): The number of columns in the embedding. See note the notes in `Embeddings` for details. (This parameter will be
            ignored when `load_file` is used.)
        n_jumps (int): Number of random jumps from node to node.
        max_per_epoch (int): Maximum number Number of consecutive random jumps without randomly jumping outside the edges. Note
            that normal random jumps are not going to explore outside a connected component.
        learn_step (float): The size of the learning step elements get approached or moved away. Units are hexadecimal degrees in along
            an ellipse.
        bidirectional (bool): Should the changes apply only to the elements of first column (False) or to both.
        load_file (str): (optional) The full path to a binary file containing a serialized GraphEmbedding object. This file must be created
            using GraphEmbedding.save().
    """

    FILE_HEAD = "mercury.graph.GraphEmbedding.1.0"
    FILE_END = "end"

    def __init__(
        self,
        dimension=None,
        n_jumps=None,
        max_per_epoch=None,
        learn_step=3,
        bidirectional=False,
        load_file=None,
    ):
        """GraphEmbedding class constructor"""
        if load_file is None and (dimension is None or n_jumps is None):
            raise ValueError(
                "Parameters dimension and n_jumps are required when load_file is None"
            )

        self.dimension = dimension
        self.n_jumps = n_jumps
        self.max_per_epoch = max_per_epoch
        self.learn_step = learn_step
        self.bidirectional = bidirectional
        self.load_file = load_file

        if self.load_file is not None:
            self._load(self.load_file)
            return

    def __getitem__(self, arg):
        """
        Method to access rows in the embedding by ID.

        Args:
            arg (same as node ids in the graph): A node ID in the graph

        Returns:
            (numpy.matrix): A numpy matrix of one row

        """
        return self.embeddings_.embeddings_matrix_[self.node_ids.index(arg)]

    def fit(self, g: Graph):
        """
        Train the embedding by doing random walks.

        Args:
            g (mercury.graph Graph asset): A `mercury.graph` Graph object. The embedding will be created so that each row in the embedding maps
                a node ID in g.

        Returns:
            (self): Fitted self (or raises an error)

        This does a number of random walks starting from a random node and selecting the edges with a probability that is proportional to
        the weight of the edge. If the destination node also has outgoing edges, the next step will start from it, otherwise, a new random
        node will be selected. The edges visited (concordant pairs) will get some reinforcement in the embedding while a randomly selected
        non-existent edges will get divergence instead (discordant pairs).

        Internally, this stores the node IDS of the node visited and calls Embeddings.fit() to transfer the structure to the embedding.
        Of course, it can be called many times on the same GraphEmbedding.

        """

        self.node_ids = list(g.networkx.nodes)

        j_matrix = nx.adjacency_matrix(g.networkx)

        N = j_matrix.shape[1]
        M = j_matrix.nnz

        self.r_ini = np.zeros(N, dtype=int)
        self.r_len = np.zeros(N, dtype=int)
        self.r_sum = np.zeros(N, dtype=float)
        self.r_col = np.zeros(M, dtype=int)
        self.r_wgt = np.zeros(M, dtype=float)

        i = 0
        for r in range(N):
            self.r_ini[r] = i

            i_col = j_matrix[[r], :].nonzero()[1]
            L = len(i_col)

            self.r_len[r] = L

            for k in range(L):
                c = i_col[k]
                w = j_matrix[r, c]

                self.r_sum[r] += w
                self.r_col[i] = c
                self.r_wgt[i] = w

                i += 1

        self.TotW = sum(self.r_sum)

        converge, diverge = _random_walks(
            self.r_ini,
            self.r_len,
            self.r_sum,
            self.r_col,
            self.r_wgt,
            self.TotW,
            self.n_jumps,
            self.max_per_epoch if self.max_per_epoch is not None else self.n_jumps,
        )

        self.embeddings_ = Embeddings(
            dimension=self.dimension,
            num_elements=len(self.node_ids),
            learn_step=self.learn_step,
            bidirectional=self.bidirectional,
        )
        self.embeddings_.fit(converge, diverge)

        return self

    def embedding(self):
        """
        Return the internal Embeddings object.

        Returns:
            (mercury.graph.embeddings.Embeddings): The embedding which is a dense matrix of `float` that can be used with `numpy` functions.
        """
        if not hasattr(self, "embeddings_"):
            return

        return self.embeddings_

    def get_most_similar_nodes(
        self, node_id, k=5, metric="cosine", return_as_indices=False
    ):
        """
        Returns the k most similar nodes and the similarities

        Args:
            node_id (object): Id of the node that we want to search the similar nodes.
            k (int): Number of most similar nodes to return
            metric (str): metric to use as a similarity.
            return_as_indices (bool): if return the nodes as indices (False), or as node ids (True)

        Returns:
            (list): list of k most similar nodes and list of similarities of the most similar nodes
            (DataFrame): A list of k most similar nodes as a `pd.DataFrame[word: string, similarity: double]`
        """
        node_index = self.node_ids.index(node_id)

        ordered_indices, ordered_similarities = (
            self.embeddings_.get_most_similar_embeddings(node_index, k, metric)
        )

        if not return_as_indices:
            nodes = list(np.array(self.node_ids)[ordered_indices])
        else:
            nodes = list(ordered_indices)

        return pd.DataFrame({"word": nodes, "similarity": ordered_similarities})

    def save(self, file_name, save_embedding=False):
        """
        Saves a GraphEmbedding to a compressed binary file with or without the embedding itself. It saves the graph's node names
        and the adjacency matrix as a sparse matrix.

        Args:
            file_name (str): The name of the file to which the GraphEmbedding will be saved.
            save_embedding (bool): Since the embedding can be big and, if not trained, it is just a matrix of uniform random numbers it is
                possible avoiding saving it. In case it is not saved, loading the file will create a new random embedding. This parameter
                controls if the embedding is saved or not (the default value).
        """
        with bz2.BZ2File(file_name, "w") as f:
            pickle.dump(GraphEmbedding.FILE_HEAD, f)
            pickle.dump(save_embedding, f)
            pickle.dump(self.embeddings_.dimension, f)

            pickle.dump(self.node_ids, f)

            np.save(f, self.r_ini)
            np.save(f, self.r_len)
            np.save(f, self.r_sum)
            np.save(f, self.r_col)
            np.save(f, self.r_wgt)

            pickle.dump(self.TotW, f)

            if save_embedding:
                np.save(f, self.embeddings_.embeddings_matrix_)

            pickle.dump(GraphEmbedding.FILE_END, f)

    def _load(self, file_name):
        """
        This method is internal and should not be called directly. Use the constructor's `load_file` argument instead.
        E.g., `ge = GraphEmbedding(load_file = 'some/stored/embedding')`
        """
        with bz2.BZ2File(file_name, "r") as f:
            head = pickle.load(f)

            if head != GraphEmbedding.FILE_HEAD:
                raise ValueError("Unsupported file format!")

            has_emb = pickle.load(f)
            dimension = pickle.load(f)

            self.node_ids = pickle.load(f)

            self.r_ini = np.load(f)
            self.r_len = np.load(f)
            self.r_sum = np.load(f)
            self.r_col = np.load(f)
            self.r_wgt = np.load(f)

            self.TotW = pickle.load(f)

            self.embeddings_ = Embeddings(dimension, len(self.node_ids))

            if has_emb:
                self.embeddings_.embeddings_matrix_ = np.load(f)

            end = pickle.load(f)

            if end != GraphEmbedding.FILE_END:
                raise ValueError("Unsupported file format!")



"""
Distributed Louvain Algorithm for Community Detection
-----------------------------------------------------
This module constitutes a PySpark implementation of the Louvain algorithm for
community detection. The algorithm aims to find the partition of a graph that
yields the maximum modularity.
"""

from mercury.graph.core.base import BaseClass
from mercury.graph.core import Graph
from mercury.graph.core.spark_interface import pyspark_installed

if pyspark_installed:
    from pyspark.sql import DataFrame, Window, functions as F

from typing import Union


class LouvainCommunities(BaseClass):
    """
    Class that defines the functions that run a PySpark implementation of the
    Louvain algorithm to find the partition that maximizes the modularity of an
    undirected graph (as in [^1]).

    This version of the algorithm differs from [^1] in that the reassignment of
    nodes to new communities is calculated in parallel, not sequentially. That is,
    all nodes are reassigned at the same time and conflicts (i.e., 1 -> C2 and
    2 -> C1) are resolved with a simple tie-breaking rule. This version also
    introduces the resolution parameter _gamma_, as in [^2].

    Contributed by Arturo Soberon Cedillo, Jose Antonio Guzman Vazquez and
    Isaac Dodanim Hernandez Garcia.

    [^1]:
        Blondel V D, Guillaume J-L, Lambiotte R and Lefebvre E (2008). Fast
        unfolding of communities in large networks. Journal of Statistical
        Mechanics: Theory and Experiment, 2008.
        <https://doi.org/10.1088/1742-5468/2008/10/p10008>

    [^2]:
        Aynaud T, Blondel V D, Guillaume J-L and Lambiotte R (2013). Multilevel
        local optimization of modularity. Graph Partitioning (315--345), 2013.

    Args:
        min_modularity_gain (float):
            Modularity gain threshold between each pass. The algorithm
            stops if the gain in modularity between the current pass
            and the previous one is less than the given threshold.

        max_pass (int):
            Maximum number of passes.

        max_iter (int):
            Maximum number of iterations within each pass.

        resolution (float):
            The resolution parameter _gamma_. Its value
            must be greater or equal to zero. If resolution is less than 1,
            modularity favors larger communities, while values greater than 1
            favor smaller communities.

        all_partitions (bool, optional):
            If True, the function will return all the partitions found at each
            step of the algorithm (i.e., pass0, pass1, pass2, ..., pass20). If
            False, only the last (and best) partition will be returned.

        verbose (bool, optional):
            If True, print progress information during the Louvain algorithm
            execution. Defaults to True.
    """

    def __init__(
        self,
        min_modularity_gain=1e-03,
        max_pass=2,
        max_iter=10,
        resolution: Union[float, int] = 1,
        all_partitions=True,
        verbose=True,
    ):
        self.min_modularity_gain = min_modularity_gain
        self.max_pass = max_pass
        self.max_iter = max_iter
        self.resolution = resolution
        self.all_partitions = all_partitions
        self.verbose = verbose

        # Check resolution
        if resolution < 0:
            exceptionMsg = f"Resolution value is {resolution} and cannot be < 0."
            raise ValueError(exceptionMsg)

    def __str__(self):
        base_str = super().__str__()

        # Check if the object has been fitted (fitting creates the `labels_` attribute)
        if hasattr(self, "labels_"):
            extra_str = [
                f"",
                f"Cluster assignments are available in attribute `labels_`",
                f"Modularity: {self.modularity_}",
            ]
            return "\n".join([base_str] + extra_str)
        else:
            return base_str

    def fit(self, g: Graph):
        """
        Args:
            g (Graph): A mercury graph structure.

        Returns:
            (self): Fitted self (or raises an error).
        """
        edges = g.graphframe.edges

        # Verify edges input
        self._verify_data(
            df=edges,
            expected_cols_grouping=["src", "dst"],
            expected_cols_others=["weight"],
        )

        # Init dataframe to be returned
        ret = (
            edges.selectExpr("src as id")
            .unionByName(edges.selectExpr("dst as id"))
            .distinct()
            .withColumn("pass0", F.row_number().over(Window.orderBy("id")))
        ).checkpoint()

        # Convert edges to anonymized src's and dst's
        edges = (
            edges.selectExpr("src as src0", "dst as dst0", "weight")
            .join(other=ret.selectExpr("id as src0", "pass0 as src"), on="src0")
            .join(other=ret.selectExpr("id as dst0", "pass0 as dst"), on="dst0")
            .select("src", "dst", "weight")
        ).checkpoint()

        # Calculate m and initialize modularity
        m = self._calculate_m(edges)
        modularity0 = -1.0

        # Begin pass
        canPass, _pass = True, 0
        while canPass:

            # Declare naive partition
            p1 = (
                edges.selectExpr("src as id")
                .unionByName(edges.selectExpr("dst as id"))
                .distinct()
                .withColumn("c", F.col("id"))
            )

            # Begin iterations within pass
            canIter, _iter = True, 0
            # Carry reference to previously cached p2 to call unpersist()
            prev_p2 = None
            while canIter:

                if _iter >= self.max_iter:
                    break

                # Print progress
                if self.verbose:
                    print(f"Starting Pass {_pass} Iteration {_iter}.")

                # Create new partition and check if movements were made
                p2 = self._reassign_all(edges, p1)
                # Break complex lineage caused by loops first
                p2 = p2.checkpoint()
                p2.cache()

                canIter = len(p2.where("cx != cj").take(1)) > 0
                if canIter:
                    p1 = p2.selectExpr("id", "cj as c")
                if prev_p2 is not None:
                    prev_p2.unpersist()
                prev_p2 = p2
                _iter += 1

            # Calculate new modularity and update pass counter
            modularity1 = self._calculate_modularity(edges=edges, partition=p1, m=m)

            # Declare stopping criterion and update old modularity
            canPass = (modularity1 - modularity0 > self.min_modularity_gain) and (
                _pass < self.max_pass
            )
            modularity0 = modularity1

            self.modularity_ = modularity0

            # Update ret and compress graph
            if canPass:
                ret = ret.join(
                    other=p1.selectExpr(f"id as pass{_pass}", f"c as pass{_pass + 1}"),
                    on=f"pass{_pass}",
                ).checkpoint()

                edges = (
                    self._label_edges(edges, p1)
                    .select("cSrc", "cDst", "weight")
                    .groupBy("cSrc", "cDst")
                    .agg(F.sum("weight").alias("weight"))
                    .selectExpr("cSrc as src", "cDst as dst", "weight")
                ).checkpoint()

            prev_p2.unpersist()
            _pass += 1

        # Return final dataframe with sorted columns
        if self.all_partitions:

            # Return sorted columns
            cols = self._sort_passes(ret)
            ret = ret.select(cols)

        # Return final dataframe with id & community
        else:
            _last = self._last_pass(ret)
            ret = ret.selectExpr("id as node_id", f"{_last} as cluster")

        self.labels_ = ret

        return self

    def _verify_data(self, df, expected_cols_grouping, expected_cols_others):
        """Checks if `edges` meets the format expected by `LouvainCommunities`.

        Args:
            df (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

            expected_cols_grouping (list):
                A list of strings representing the columns that must be present in
                `df` to group the data.

            expected_cols_others (list):
                A list of strings representing the columns that must be present in
                `df` but are not used for grouping.
        """

        cols = df.columns
        expected_cols = expected_cols_grouping + expected_cols_others

        # Check type
        if not isinstance(df, DataFrame):
            raise TypeError("Input data must be a pyspark DataFrame.")

        # Check missing columns
        msg = "Input data is missing expected column '{}'."
        for col in expected_cols:
            if col not in cols:
                raise ValueError(msg.format(col))

        # Check for duplicates
        dup = (
            df.groupBy(*expected_cols_grouping)
            .agg(F.count(F.lit(1)).alias("count"))
            .where("count > 1")
            .count()
        )
        if dup > 0:
            raise ValueError("Data has duplicated entries.")

    def _last_pass(self, df):
        """Returns the column name of the last pass.

        Args:
            df (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the series of partitions made by
                `LouvainCommunities` (a dataframe with columns 'id', 'pass0',
                'pass1', 'pass2', 'pass3', etc.).
        """

        # Get all `passX` columns as list
        cols = [col for col in df.columns if "pass" in col]

        # Get last pass as int
        _max = max([int(col.split("pass")[1]) for col in cols])

        # Return last pass as string
        return f"pass{_max}"

    def _label_degrees(self, edges, partition):
        """
        This function uses the edges of a graph to calculate the weighted degrees
        of each node and joins the result with the partition passed by the user.

        Args:
            edges (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

            partition (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the partition of an undirected
                graph (i.e., a table that indicates the community that each node
                belongs to). The dataframe must have columns `id` (indicating each
                node's ID) and `c` (indicating each node's assigned community).

        Returns:
            (Dataframe):
                This function returns a dataframe with columns `id` (representing the ID
                of each node in the graph), `c` (representing each node's community) and
                `degree` (representing each node's degree).
        """

        # Get id, community and weighted degree
        ret = (
            partition.join(
                # Unite sources and destinations to avoid double join
                other=(
                    edges.selectExpr("src as id", "weight")
                    .unionByName(edges.selectExpr("dst as id", "weight"))
                    .groupBy("id")
                    .agg(F.sum("weight").alias("degree"))
                ),
                on="id",
                how="inner",
            )
            .select("id", "c", "degree")
            .checkpoint()
        )

        return ret

    def _label_edges(self, edges, partition):
        """This function uses `partition` to add two columns to `edges`. The added
        columns `cSrc` and `cDst` indicate the community that the source and
        destination nodes belong to.

            Args:
            edges (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

            partition (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the partition of an undirected
                graph (i.e., a table that indicates the community that each node
                belongs to). The dataframe must have columns `id` (indicating each
                node's ID) and `c` (indicating each node's assigned community).

        Returns:
            (pyspark.sql.dataframe.DataFrame):
                This function returns `edges` with two additional columns: the community
                that the source node belongs to (`cSrc`) and the community that the
                destination node belongs to (`cDst`).
        """

        # Get communities
        ret = (
            edges
            # Start off with src, dst and weight
            .select("src", "dst", "weight")
            # Source destination
            .join(
                other=partition.selectExpr("id as src", "c as cSrc"),
                on="src",
                how="left",
            )
            # Destination community
            .join(
                other=partition.selectExpr("id as dst", "c as cDst"),
                on="dst",
                how="left",
            ).checkpoint()
        )

        return ret

    def _calculate_m(self, edges) -> int:
        """Get the weighted size of an undirected graph (where $m$ is
        defined as $m = \\frac{1}{2} \\sum_{ij} A_{ij}$)).

        Args:
            edges (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

        Returns:
            (int): Returns the weighted size of the graph.
        """

        m = edges.select(F.sum("weight")).collect()[0][0]

        return int(m)

    def _calculate_modularity(self, edges, partition, m=None) -> float:
        """This function calculates the modularity of a partition.

        Args:
            edges (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

            partition (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the partition of an undirected
                graph (i.e., a table that indicates the community that each node
                belongs to). The dataframe must have columns `id` (indicating each
                node's ID) and `c` (indicating each node's assigned community).

            m (int):
                The weighted size of the graph (the output of `_get_m()`).

        Returns:
            (float):
                Bound between -1 and 1 representing the modularity of a
                partition. The output may exceed these bounds depending on the value of
                `resolution` (which is set to 1 by default).
        """

        # Calculate m (if necessary) and norm
        m = self._calculate_m(edges) if m is None else m
        norm = 1 / (2 * m)

        # Declare basic inputs
        labeledEdges = self._label_edges(edges, partition)
        labeledDegrees = self._label_degrees(edges, partition)

        # Get term on LHS
        k_in = (labeledEdges.where("cSrc = cDst").select(F.sum("weight"))).collect()[0][
            0
        ]

        # Handle NoneType
        k_in = 0 if k_in is None else k_in

        # Get term on RHS
        k_out = (
            labeledDegrees.groupby("c")
            .agg(F.sum("degree").alias("kC"))
            .selectExpr(f"{self.resolution} * sum(kC * kC)")
        ).collect()[0][0]

        # Return modularity
        return (k_in / m) - (norm**2 * float(k_out))

    def _reassign_all(self, edges, partition, m=None):
        """This function simultaneously reassigns all the nodes in a graph to their
        corresponding optimal neighboring communities.

        Args:
            edges (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the edges of an undirected graph.
                It must have `src` and `dst` as its columns. The user may also
                specify the weight of each edge via the additional `weight` column
                (optional).

            partition (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the partition of an undirected
                graph (i.e., a table that indicates the community that each node
                belongs to). The dataframe must have columns `id` (indicating each
                node's ID) and `c` (indicating each node's assigned community).

            m (int):
                The weighted size of the graph (the output of `getM()`).

        Returns:
            (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe with the same number of rows as there are vertices.
                Columns `cx` and `cj` represent each node's current and optimal
                neighboring community (accordingly).
        """

        # Calculate m if necessary
        m = self._calculate_m(edges) if m is None else m

        # Label edges and degrees here to avoid long lineages
        labeledDegrees = self._label_degrees(edges, partition)
        labeledEdges = self._label_edges(edges, partition)

        # Add sum(ki) for i in C to labeledDegrees
        dq = (
            labeledDegrees.withColumn(
                "cx_sum_ki", F.sum("degree").over(Window.partitionBy("c"))
            )
            # Get sum(Aix) for i in Cx\{x}
            .join(
                other=(
                    labeledEdges.where("(src != dst) and (cSrc = cDst)")
                    .selectExpr("src as id", "weight")
                    .unionByName(
                        labeledEdges.where("(src != dst) and (cSrc = cDst)").selectExpr(
                            "dst as id", "weight"
                        )
                    )
                    .groupBy("id")
                    .agg(F.sum("weight").alias("cx_sum_aix"))
                ),
                on="id",
                how="left",
            )
            # Get sum(Aix) for i in Cj (relationship 1:J)
            .join(
                other=(
                    labeledEdges.where("cSrc != cDst")
                    .selectExpr("src as id", "cDst as cj", "weight")
                    .unionByName(
                        labeledEdges.where("cSrc != cDst").selectExpr(
                            "dst as id", "cSrc as cj", "weight"
                        )
                    )
                    .groupBy("id", "cj")
                    .agg(F.sum("weight").alias("cj_sum_aix"))
                ),
                on="id",
                how="left",
            )
            # Get sum(ki) for i in Cj
            .join(
                other=(
                    labeledDegrees.withColumnRenamed("c", "cj")
                    .groupBy("cj")
                    .agg(F.sum("degree").alias("cj_sum_ki"))
                ),
                on="cj",
                how="left",
            )
            # Calculate modularity change of each possible switch (Cx -> {x} -> Cj)
            .withColumn(
                "mdq",
                F.coalesce("cj_sum_aix", F.lit(0))
                - F.coalesce("cx_sum_aix", F.lit(0))
                - (
                    F.col("degree")
                    / F.lit(2 * m)
                    * (F.col("cj_sum_ki") - F.col("cx_sum_ki") + F.col("degree"))
                ),
            )
            # Rank mdq(x) in descending order
            .select(
                F.col("id"),
                F.col("c"),
                F.coalesce("cj", F.col("c")).alias("cj"),  # Trapped nodes: Cx == Cj
                F.col("mdq"),
                F.row_number()
                .over(Window.partitionBy("id").orderBy(F.desc("mdq")))
                .alias("mdq_rank"),
            )
            # Keep best (or first) change
            .where(F.col("mdq_rank") == 1)
        )

        # Break symmetric swaps (only in first iteration?)
        dq = (
            dq.withColumn(
                "sym_rank",
                F.row_number().over(
                    Window.partitionBy(
                        F.sort_array(F.array(F.col("c"), F.col("cj")))
                    ).orderBy(F.desc("mdq"))
                ),
            )
            # Select best switch (cStar) and break symmetric swaps with sym_rank
            .withColumn(
                "cStar",
                F.when(
                    ((F.col("mdq") > F.lit(1e-04)) & (F.col("sym_rank") == 1)),
                    F.col("cj"),
                ).otherwise(F.col("c")),
            ).selectExpr("id", "c as cx", "cStar as cj")
        )

        return dq

    def _sort_passes(self, res) -> list:
        """Takes the output of `LouvainCommunities` and returns a list containing
        its columns ordered by their integer part in ascending order.
        For example, if the columns returned by `LouvainCommunities are
        `['pass2', 'id', 'pass1', 'pass0']`, this function will turn the list to
        `['id', 'pass0', 'pass1', 'pass2']`.
        This function also supports cases where `max_pass > 10`.

        Args:
            res (pyspark.sql.dataframe.DataFrame):
                A pyspark dataframe representing the output of `LouvainCommunities`.
                `res` must have columns 'id', 'pass0', 'pass1', 'pass2', etc.
        """

        # Get pass-columns and sort them by their integer part
        cols = [col for col in res.columns if "pass" in col]
        ints = sorted([int(col.replace("pass", "")) for col in cols])
        cols_sorted = ["id"] + ["pass" + str(i) for i in ints]

        return cols_sorted



import logging

from mercury.graph.core import Graph
from mercury.graph.core.base import BaseClass

from mercury.graph.core.spark_interface import (
    SparkInterface,
    pyspark_installed,
    graphframes_installed,
)

if pyspark_installed:
    import pyspark.sql.functions as f
    from pyspark.sql import Window

    from pyspark.sql.types import ArrayType
    from pyspark.sql.types import StringType

    from pyspark.ml.feature import Word2Vec
    from pyspark.ml.feature import Word2VecModel

    def _add_new_element_2(acc_path, next_node):
        return (
            acc_path + [next_node]
            if ((next_node is not None) and (acc_path is not None))
            else acc_path
        )

    udf_select_element_2 = f.udf(_add_new_element_2, ArrayType(StringType()))

    if graphframes_installed:
        from graphframes import GraphFrame
        from graphframes.lib import AggregateMessages


class SparkNode2Vec(BaseClass):

    def __init__(
        self,
        dimension=None,
        sampling_ratio=1.0,
        num_epochs=10,
        num_paths_per_node=1,
        batch_size=1000000,
        w2v_max_iter=1,
        w2v_num_partitions=1,
        w2v_step_size=0.025,
        w2v_min_count=5,
        path_cache=None,
        use_cached_rw=False,
        n_partitions_cache=10,
        load_file=None,
    ):
        """
        Create or reload a SparkNode2Vec embedding mapping the nodes of a graph.

        Args:
            dimension (int): The number of columns in the embedding. See note the notes in `Embeddings` for details. (This parameter will be
                ignored when `load_file` is used.)
            sampling_ratio (float): The proportion from the total number of nodes to be used in parallel at each step (whenever possible).
            num_epochs (int): Number of epochs. This is the total number of steps the iteration goes through. At each step, sampling_ratio
                times the total number of nodes paths will be computed in parallel.
            num_paths_per_node (int): The amount of random walks to source from each node.
            batch_size (int): This forces caching the random walks computed so far and breaks planning each time this number of epochs
                is reached. The default value is a high number to avoid this entering at all. In really large jobs, you may want to
                set this parameter to avoid possible overflows even if it can add some extra time to the process. Note that with a high
                number of epochs and nodes resource requirements for the active part of your random walks can be high. This allows to
                "cache a continue" so to say.
            w2v_max_iter (int): This is the Spark Word2Vec parameter maxIter, the default value is the original default value.
            w2v_num_partitions (int): This is the Spark Word2Vec parameter numPartitions, the default value is the original default value.
            w2v_step_size (float): This is the Spark Word2Vec parameter stepSize, the default value is the original default value.
            w2v_min_count (int): This is the Spark Word2Vec parameter minCount, the default value is the original default value (5). Is the
                minimum number of times that a node has to appear to generate an embedding.
            path_cache (str): Folder where random walks will be stored, the default value is None which entails that random walks will not
                be stored.
            use_cached_rw (bool): Flag that indicates if random walks should be read from disk (hence, they will not be computed again).
                Setting this parameter to True requires a valid path_cache.
            n_partitions_cache (int): Number of partitions that will be used when storing the random walks, to optimize read access.
                The default value is 10.
            load_file (str): (optional) The full path to a parquet file containing a serialized SparkNode2Vec object. This file must be created
                using SparkNode2Vec.save().
        """
        self.dimension = dimension
        self.sampling_ratio = sampling_ratio
        self.num_epochs = num_epochs
        self.num_paths_per_node = num_paths_per_node
        self.batch_size = batch_size
        self.w2v_max_iter = w2v_max_iter
        self.w2v_num_partitions = w2v_num_partitions
        self.w2v_step_size = w2v_step_size
        self.w2v_min_count = w2v_min_count
        self.path_cache = path_cache
        self.use_cached_rw = use_cached_rw
        self.n_partitions_cache = n_partitions_cache
        self.load_file = load_file

        if self.load_file is not None:
            self._load(self.load_file)
            return

    def __str__(self):
        base_str = super().__str__()

        # Check if the object has been fitted (fitting creates `node2vec_` model)
        if hasattr(self, "node2vec_"):
            extra_str = [
                f"",
                f"Random walk paths are available in attribute `paths_`.",
                f"Spark's Word2Vec model fitted on paths_ is available in attribute `node2vec_` through method `model()`.",
            ]
            return "\n".join([base_str] + extra_str)
        else:
            return base_str

    def fit(self, G: Graph):
        """
        Train the embedding by doing random walks.

        Random walk paths are available in attribute `paths_`.

        Args:
            G (mercury.graph.core.Graph): A `mercury.graph` Graph object. The embedding will be created so that each row in the embedding maps
                a node ID in G. (This parameter will be ignored when `load_file` is used.)

        Returns:
            (self): Fitted self (or raises an error)
        """

        if self.path_cache is None:
            if self.use_cached_rw:
                logging.warning(
                    "Wrong options (use_cached_rw and no path_cache). "
                    "Paths will be recomputed."
                )
            self.use_cached_rw = False

        if not self.use_cached_rw:
            paths = (
                self._run_rw(G)
                .withColumn("size", f.size("random_walks"))
                .where(f.col("size") > 1)
                .drop("size")
            )

            if self.path_cache is not None:
                (
                    paths.repartition(self.n_partitions_cache)
                    .write.mode("overwrite")
                    .parquet("%s/block=0" % self.path_cache)
                )

            if self.num_paths_per_node > 1:
                for block_id in range(1, self.num_paths_per_node):
                    new_paths = (
                        self._run_rw(G)
                        .withColumn("size", f.size("random_walks"))
                        .where(f.col("size") > 1)
                        .drop("size")
                    )
                    if self.path_cache is None:
                        paths = paths.unionByName(new_paths)
                    else:
                        (
                            new_paths.repartition(self.n_partitions_cache)
                            .write.mode("overwrite")
                            .parquet("%s/block=%d" % (self.path_cache, block_id))
                        )
                        # With this, we clear the persisted dataframe
                        new_paths.unpersist()

        if self.path_cache is None:
            self.paths_ = paths.persist()
        else:
            self.paths_ = (
                SparkInterface()
                .read_parquet(self.path_cache)
                .drop("block")
                .repartition(self.n_partitions_cache)
                .persist()
            )

        w2v = Word2Vec(
            vectorSize=self.dimension,
            maxIter=self.w2v_max_iter,
            numPartitions=self.w2v_num_partitions,
            stepSize=self.w2v_step_size,
            inputCol="random_walks",
            outputCol="model",
            minCount=self.w2v_min_count,
        )

        self.node2vec_ = w2v.fit(self.paths_)

        return self

    def embedding(self):
        """
        Return all embeddings.

        Returns:
            (DataFrame): All embeddings as a `DataFrame[word: string, vector: vector]`.
        """
        if not hasattr(self, "node2vec_"):
            return

        return self.node2vec_.getVectors()

    def model(self):
        """
        Returns the Spark Word2VecModel object.

        Returns:
            (pyspark.ml.feature.Word2VecModel): The Spark Word2VecModel of the embedding to use its API directly.
        """
        if not hasattr(self, "node2vec_"):
            return

        return self.node2vec_

    def get_most_similar_nodes(self, node_id, k=5):
        """
        Returns the k most similar nodes and a similarity measure.

        Args:
            node_id (str): Id of the node we want to search.
            k (int): Number of most similar nodes to return

        Returns:
            (DataFrame): A list of k most similar nodes (using cosine similarity) as a `DataFrame[word: string, similarity: double]`
        """
        if not hasattr(self, "node2vec_"):
            return

        return self.node2vec_.findSynonyms(node_id, k)

    def save(self, file_name):
        """
        Saves the internal Word2VecModel to a human-readable (JSON) model metadata as a Parquet formatted data file.

        The model may be loaded using SparkNode2Vec(load_file='path/file')

        Args:
            file_name (str): The name of the file to which the Word2VecModel will be saved.
        """
        if not hasattr(self, "node2vec_"):
            return

        return self.node2vec_.save(file_name)

    def _load(self, file_name):
        """
        This method is internal and should not be called directly. Use the constructor's `load_file` argument instead.
        E.g., `snv = SparkNode2Vec(load_file = 'some/stored/embedding')`
        """

        self.node2vec_ = Word2VecModel.load(file_name)

    def _start_rw(self, G: Graph):
        aux_vert = (
            G.graphframe.vertices.groupBy(f.col("id"))
            .agg(f.collect_list(f.col("id")).alias("tmp_rw_aux_acc_path"))
            .withColumn("tmp_rw_init_p", f.rand())
            .withColumn(
                "new_rw_acc_path",
                f.when(
                    f.col("tmp_rw_init_p") <= self.sampling_ratio,
                    f.col("tmp_rw_aux_acc_path"),
                ).otherwise(f.lit(None)),
            )
            .withColumn("new_rw_curr_id", f.col("id"))
            .drop("tmp_rw_aux_acc_path", "tmp_rw_init_p")
        ).filter(f.col("new_rw_acc_path").isNotNull())

        w_ind = (
            Window.partitionBy("src")
            .orderBy("dst")
            .rangeBetween(Window.unboundedPreceding, 0)
        )
        w_ind_edge = Window.partitionBy("src")

        df_edges_cumsum = G.graphframe.edges.withColumn(
            "new_rw_cum_sum", f.sum("weight").over(w_ind)
        ).withColumn("new_rw_total_sum", f.sum("weight").over(w_ind_edge))

        aux_edges = (
            df_edges_cumsum.withColumn(
                "new_rw_norm_cumsum",
                f.col("new_rw_cum_sum") / f.col("new_rw_total_sum"),
            ).select(
                f.col("src"),
                f.col("dst"),
                f.col("weight"),
                f.col("new_rw_norm_cumsum"),
            )
        )

        self.gx = GraphFrame(aux_vert, aux_edges)

    def _update_state_with_next_step(self, i):
        candidate_vert_col = self.gx.vertices.columns

        candidate_vert = self.gx.vertices.withColumnRenamed(
            "new_rw_curr_id", "new_rw_prev_id"
        ).withColumn("rand_p_step", f.rand())

        candidate_vert_with_edges = candidate_vert.join(
            self.gx.edges,
            f.col("new_rw_prev_id") == f.col("src"),
            "left",
        ).filter(
            (f.col("rand_p_step") <= f.col("new_rw_norm_cumsum"))
            | (f.isnull(f.col("new_rw_norm_cumsum")))
        )

        win_steps = Window.partitionBy("id").orderBy(f.col("new_rw_norm_cumsum").asc())

        selected_next_step = candidate_vert_with_edges.withColumn(
            "row_number", f.row_number().over(win_steps)
        ).filter(f.col("row_number") <= 1)

        selected_next_step = (
            selected_next_step.withColumn(
                "new_rw_acc_path",
                udf_select_element_2(f.col("new_rw_acc_path"), f.col("dst")),
            )
            .withColumn("new_rw_curr_id", f.col("new_rw_acc_path").getItem(i + 1))
            .select(candidate_vert_col)
        )

        return selected_next_step

    def _run_rw(self, G: Graph):
        self._start_rw(G)

        for i in range(self.num_epochs):

            aux_vert = self._update_state_with_next_step(i)
            aux_vert = aux_vert.checkpoint()
            self.gx = GraphFrame(aux_vert, self.gx.edges)

            if (i + 1) % self.batch_size == 0:
                old_aux_vert = aux_vert
                aux_vert = AggregateMessages.getCachedDataFrame(aux_vert)
                old_aux_vert.unpersist()

        paths = self.gx.vertices.select(
            f.col("new_rw_acc_path").alias("random_walks")
        ).filter(f.col("random_walks").isNotNull())

        self.gx.vertices.unpersist()
        self.gx.edges.unpersist()

        return paths



import importlib.util

spec = importlib.util.find_spec('numba')


if spec is None:
    def njit(f):
        """
        This decorator does nothing, it is a fallback in case numba is not installed.
        """

        def wrapper(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapper

    graph_i4 = int
    numba_installed = False


else:
    from numba import njit
    from numba.types import i4 as graph_i4
    numba_installed = True



import os, shutil


# Move tutorials inside mercury.graph before packaging
if os.path.exists('tutorials'):
    shutil.move('tutorials', 'mercury/graph/tutorials')


from setuptools import setup, find_packages


setup_args = dict(
	name				 = 'mercury-graph',
	packages			 = find_packages(include = ['mercury*', 'tutorials*'], exclude = ['docker', 'unit_tests']),
	scripts				 = ['test.sh'],
	include_package_data = True,
	package_data		 = {'mypackage': ['tutorials/*', 'tutorials/data/*']}
)

setup(**setup_args)



from typing import Union, List
import itertools

from mercury.graph.core import Graph
from mercury.graph.core.spark_interface import (
    SparkInterface,
    pyspark_installed,
    graphframes_installed,
)
from mercury.graph.core.base import BaseClass

if pyspark_installed:
    import pyspark
    from pyspark.sql import Row
    import pyspark.sql.functions as f
    from pyspark.sql.types import StructField, StructType, StringType, ArrayType

    def _add_new_nodes(acc_nodes, new_nodes):
        if (new_nodes is None) or (len(new_nodes) == 0):
            return acc_nodes
        else:
            new_nodes = list(itertools.chain.from_iterable(new_nodes))
            if acc_nodes is None:
                return new_nodes
            return list(set(acc_nodes + new_nodes))

    udf_add_new_nodes = f.udf(_add_new_nodes, ArrayType(StringType()))

    if graphframes_installed:
        from graphframes import GraphFrame
        from graphframes.lib import AggregateMessages as AM


class SparkSpreadingActivation(BaseClass):
    """
    This class is a model that represents a “word-of-mouth” scenario where a node influences his neighbors, from where
    the influence spreads to other neighbors, and so on.

    At the end of the diffusion process, we inspect the amount of influence received by each node. Using a
    threshold-based technique, a node that is currently not influenced can be declared to be a potential future one,
    based on the influence that has been accumulated.

    The diffusion model is based on Spreading Activation (SPA) techniques proposed in cognitive psychology
    and later used for trust metric computations. For more details, please see paper entitled
    ["Social Ties and their Relevance to Churn in Mobile Telecom Networks"](https://pdfs.semanticscholar.org/3275/3d80adb5ec2d4a974b5d1a872e2c957b263b.pdf)

    Args:
        attribute (str): Column name which will store the amount of influence spread
        spreading_factor (float): Percentage of influence to distribute. Low values favor influence proximity to the source of injection,
            while high values allow the influence to also reach nodes which are further away. It must be a value in the range (0,1).
            Default value is 0.2
        transfer_function (str): Allowed values: "weighted" or "unweighted".
            Once a node decides what fraction of energy to distribute, the next step is to decide what fraction of the energy is transferred
            to each neighbor. This is controlled by the Transfer Function. If "weighted" then the energy distributed along the directed
            edge <X,Y> depends on its relatively weight compared to the sum of weights of all outgoing edges of X. If "unweighted", then
            the energy distributed along the edge <X,Y> is independent of its relatively weight.
        steps (int): Number of steps to perform
        influenced_by (bool): if True, and extra column "influenced_by" is calculated which contains the seed nodes that have spread some
            influence to a given node. When True, the ids of the nodes cannot contain commas ",". Note that seed_nodes will have at least
            their own (remaining) influence
    """

    def __init__(
        self,
        attribute: str = "influence",
        spreading_factor: float = 0.2,
        transfer_function: str = "weighted",
        steps: int = 1,
        influenced_by: bool = False,
    ):
        self.attribute = attribute
        self.spreading_factor = spreading_factor
        self.transfer_function = transfer_function
        self.steps = steps
        self.influenced_by = influenced_by

    def fit(
        self,
        g: Graph,
        seed_nodes: Union[List, "pyspark.sql.DataFrame"],
    ):
        """
        Perform all iterations of spread_activation

        Args:
            g (mercury.graph.core.Graph): A `mercury.graph` Graph object.
            seed_nodes (Union[List, pyspark.sql.DataFrame]): Collection of nodes that are the "seed" or are the source to spread
                the influence. It must be pyspark dataframe with column 'id' or python list

        Returns:
            (self): Fitted self
        """

        # Set seed nodes which are the source of influence
        g = self._set_seed_nodes(g, seed_nodes)

        # Compute degrees
        g = self._compute_degrees(g)

        # Number of iterations specified for spread activation
        for _ in range(0, self.steps, 1):
            g = self._spread_activation_step(
                g,
            )

        # Graph with updated attributes
        self.fitted_graph_ = g
        # Influences as DataFrame
        self.influences_ = self.fitted_graph_.nodes_as_dataframe().select(
            "id", "influence"
        )

        return self

    def _set_seed_nodes(
        self,
        g: Graph,
        seed_nodes: Union[List, "pyspark.sql.DataFrame"] = None,
    ):
        """
        Set seed nodes which are the source of influence using pyspark dataframe.

        Args:
            g (mercury.graph.core.Graph): A `mercury.graph` Graph object.
            seed_nodes (Union[List, pyspark.sql.DataFrame]): Collection of nodes that are the source to spread
                the influence. It must be pyspark dataframe with column 'id' or python list.
        """

        seed_nodes_dataframe = seed_nodes

        # Convert list to dataframe
        if isinstance(seed_nodes, list):
            rdd_list = SparkInterface().spark.sparkContext.parallelize(seed_nodes)
            row_rdd_list = rdd_list.map(lambda x: Row(x))
            field_list = [StructField("id", StringType(), True)]
            schema_list = StructType(field_list)
            seed_nodes_dataframe = SparkInterface().spark.createDataFrame(
                row_rdd_list, schema_list
            )

        # Create column for influence attribute containing 1's
        seed_nodes_dataframe = seed_nodes_dataframe.withColumn(
            self.attribute, f.lit(1.0)
        )
        self.seed_nodes_ = seed_nodes_dataframe

        # Merge to original vertices of graph
        orig_vertices = g.graphframe.vertices.select("id")
        orig_edges = g.graphframe.edges
        new_vertices = orig_vertices.join(
            seed_nodes_dataframe, "id", "left_outer"
        ).na.fill(0)

        # If influenced_by flag is set, then initialize the seed nodes
        if self.influenced_by:
            new_vertices = new_vertices.withColumn(
                "influenced_by",
                f.when(
                    new_vertices[self.attribute] == 1,
                    f.split(new_vertices["id"], pattern=","),
                ).otherwise(f.array().cast("array<string>")),
            )

        # Update graph
        return Graph(GraphFrame(new_vertices, orig_edges))

    def _compute_degrees(self, g: Graph):
        """
        Compute weighted and unweighted in and out degrees in graph. Re-declares graph to add the following
        attributes: inDegree, outDegree, w_inDegree, w_outDegree.

        Args:
            g: graphframe object, network
        """
        g_vertices = g.graphframe.vertices
        g_edges = g.graphframe.edges

        # Get unweighted degrees
        indeg = g.graphframe.inDegrees
        outdeg = g.graphframe.outDegrees

        # Get weighted degrees
        w_indeg = (
            g_edges.groupby("dst").agg(f.sum("weight").alias("w_inDegree"))
        ).selectExpr("dst as id", "w_inDegree as w_inDegree")
        w_outdeg = (
            g_edges.groupby("src").agg(f.sum("weight").alias("w_outDegree"))
        ).selectExpr("src as id", "w_outDegree as w_outDegree")

        # Update vertices attribute
        new_v = g_vertices.join(indeg, "id", "left_outer")
        new_v = new_v.join(outdeg, "id", "left_outer")
        new_v = new_v.join(w_indeg, "id", "left_outer")
        new_v = new_v.join(w_outdeg, "id", "left_outer")
        new_v = new_v.na.fill(0)

        # Update graph
        return Graph(GraphFrame(new_v, g_edges))

    def _spread_activation_step(self, g: Graph):
        """
        One step in the spread activation model.

        Args:
            g: graphframe object, network

        Returns:
            (Graphframe): new network with updated new calculation of attribute in vertices
        """

        # Pass influence/message to neighboring nodes (weighted/unweighted option)
        if self.transfer_function == "unweighted":
            msg_to_src = (AM.src[self.attribute] / AM.src["outDegree"]) * (
                1 - self.spreading_factor
            )
            msg_to_dst = f.when(
                AM.dst["outDegree"] != 0,
                (
                    (AM.src[self.attribute] / AM.src["outDegree"])
                    * self.spreading_factor
                ),
            ).otherwise(
                ((1 / AM.dst["inDegree"]) * AM.dst[self.attribute])
                + (
                    (AM.src[self.attribute] / AM.src["outDegree"])
                    * self.spreading_factor
                )
            )

        elif self.transfer_function == "weighted":
            weight = AM.edge["weight"] / AM.src["w_outDegree"]
            msg_to_src = (AM.src[self.attribute] / AM.src["outDegree"]) * (
                1 - self.spreading_factor
            )
            msg_to_dst = f.when(
                AM.dst["outDegree"] != 0,
                ((AM.src[self.attribute]) * (self.spreading_factor * weight)),
            ).otherwise(
                ((1 / AM.dst["inDegree"]) * AM.dst[self.attribute])
                + ((AM.src[self.attribute]) * (self.spreading_factor * weight))
            )

        # Aggregate messages
        agg = g.graphframe.aggregateMessages(
            f.sum(AM.msg).alias(self.attribute),
            sendToSrc=msg_to_src,
            sendToDst=msg_to_dst,
        )

        # Create a new cached copy of the dataFrame to get new calculated attribute
        cached_new_vertices = AM.getCachedDataFrame(agg)

        if self.influenced_by:
            to_join = g.graphframe.vertices.select(
                "id",
                "inDegree",
                "outDegree",
                "w_inDegree",
                "w_outDegree",
                "influenced_by",
            )
        else:
            to_join = g.graphframe.vertices.select(
                "id", "inDegree", "outDegree", "w_inDegree", "w_outDegree"
            )
        new_cached_new_vertices = cached_new_vertices.join(to_join, "id", "left_outer")
        new_cached_new_vertices = new_cached_new_vertices.na.fill(0)

        # If influenced_by flag is set, compute new seed nodes influencing
        if self.influenced_by:
            new_cached_new_vertices = self._calculate_influenced_by(
                g, new_cached_new_vertices
            )

        # Return graph with new calculated attribute
        return Graph(GraphFrame(new_cached_new_vertices, g.graphframe.edges))

    def _calculate_influenced_by(self, g: Graph, vertices):
        influencing_msg_to_dst = AM.src["influenced_by"]
        agg_influencing = g.graphframe.aggregateMessages(
            f.collect_list(AM.msg).alias("new_influences"),
            sendToDst=influencing_msg_to_dst,
        )
        cached_new_influences = AM.getCachedDataFrame(agg_influencing)

        # Join with current nodes and concatenate new influencing nodes in "influenced_by" column
        vertices = (
            vertices.join(cached_new_influences, "id", "left_outer")
            .withColumn(
                "influenced_by",
                udf_add_new_nodes(f.col("influenced_by"), f.col("new_influences")),
            )
            .drop("new_influences")
        )

        return vertices



import inspect


class BaseClass:
    """
    Base class for common functions and utilities
    """

    def __repr__(self):
        params = ", ".join([f"{k}={v}" for k, v in self.get_params().items()])

        return f"{self.__class__.__name__}({params})"

    def __str__(self):
        params = ", ".join([f"{k}={v}" for k, v in self.get_params().items()])

        base_str = [
            f"Object of class {self.__class__.__name__}.",
            f"",
            f"Initialization parameters: {params}",
        ]

        return "\n".join(base_str)

    def get_params(self):
        """
        Extract a dictionary of the object's parameters

        Returns:
            params (dict): Dictionary of parameters ({name: value})
        """
        # Extract parameter names from the constructor
        init_signature = inspect.signature(self.__init__)
        init_param_names = sorted(
            [
                param_name
                for param_name, _ in init_signature.parameters.items()
                if param_name != "self"
            ]
        )

        # Build dictionary with parameter names and values
        params = {k: getattr(self, k) for k in init_param_names}

        return params
